"""Create utility functions for interacting with data returned from TIDAL API."""

import base64
import logging
import os
import socket
import tempfile
from contextlib import closing, contextmanager
from io import BytesIO
from pathlib import Path
from typing import Optional, Tuple, Union

from Crypto.Cipher import AES
from requests import Session

TIDAL_API_URL: str = "https://api.tidal.com/v1"
IMAGE_URL: str = "https://resources.tidal.com/images/%s.jpg"

logger = logging.getLogger(__name__)


def replace_illegal_characters(input_str: str) -> str:
    """Some characters are illegal for use as file names on Windows
    and on Unix-like systems. This function replaces any of the
    forbidden characters found in input_str with a replacement;
    mostly the empty string."""
    s = (
        input_str.replace("/", "_")
        .replace("|", "_")
        .replace(":", " -")
        .replace('"', "")
        .replace(">", "")
        .replace("<", "")
        .replace("/", "")
        .replace("\\", "")
        .replace("?", "")
        .replace(" ?", "")
        .replace("? ", "")
        .replace("*", "")
        .replace("\0", "")  # ASCII null character
    )
    return s


def download_cover_image(
    session: Session,
    cover_uuid: str,
    output_dir: Path,
    file_name: str = "cover.jpg",
    dimension: Union[int, Tuple[int]] = 1280,
) -> Optional[Path]:
    """Given a UUID that corresponds to a (JPEG) image on Tidal's servers,
    download the image file and write it as 'cover.jpeg' in the directory
    'output_dir'. Returns path to downloaded file"""
    cover_url_part: str = cover_uuid.replace("-", "/")
    if isinstance(dimension, int):
        _url: str = IMAGE_URL % f"{cover_url_part}/{dimension}x{dimension}"
    elif isinstance(dimension, tuple):
        _url: str = IMAGE_URL % f"{cover_url_part}/{dimension[0]}x{dimension[1]}"

    with session.get(url=_url, headers={"Accept": "image/jpeg"}) as r:
        if not r.ok:
            logger.warning(
                "Could not retrieve data from TIDAL resources/images URL "
                f"due to error code: {r.status_code}"
            )
            logger.debug(r.reason)
            return
        else:
            bytes_to_write = BytesIO(r.content)

    if bytes_to_write is not None:
        output_file: Path = output_dir / file_name
        bytes_to_write.seek(0)
        output_file.write_bytes(bytes_to_write.read())
        bytes_to_write.close()
        return output_file


@contextmanager
def temporary_file(suffix: str = ".mka"):
    """This context-managed function is a stand-in for
    tempfile.NamedTemporaryFile as that stdlib object experiences
    errors on Windows."""
    file_name: str = os.path.join(
        tempfile.gettempdir(), f"{os.urandom(24).hex()}{suffix}"
    )
    if not os.path.exists(file_name):
        open(file=file_name, mode="x").close()

    tf = open(file=file_name, mode="wb")
    try:
        yield tf
    finally:
        tf.close()
        os.unlink(tf.name)


def decrypt_manifest_key_id(manifest_key_id: str) -> Tuple[bytes, bytes]:
    """Given a 'keyId' value from the TIDAL API manifest response, use the
    master_key gleaned from previous projects and decrypt the audio bytes.
    This will work if the manifest specifies encryption type 'OLD_AES'.
    Returns a tuple of bytes, representing the key and nonce to use to
    decrypt the audio data."""

    # https://github.com/yaronzz/Tidal-Media-Downloader/blob/master/TIDALDL-PY/tidal_dl/decryption.py#L25
    master_key: str = "UIlTTEMmmLfGowo/UC60x2H45W6MdGgTRfo/umg4754="

    # Decode the base64 strings to ascii strings
    master_key_bytes: bytes = base64.b64decode(master_key)
    manifest_key_bytes: bytes = base64.b64decode(manifest_key_id)

    # Get the IV from the first 16 bytes of the manifest's keyId
    iv: bytes = manifest_key_bytes[:16]
    encrypted_manifest_key_bytes: bytes = manifest_key_bytes[16:]

    decryptor = AES.new(master_key_bytes, AES.MODE_CBC, iv)
    decrypted_manifest_key_bytes: bytes = decryptor.decrypt(
        encrypted_manifest_key_bytes
    )

    key, nonce = decrypted_manifest_key_bytes[:16], decrypted_manifest_key_bytes[16:24]
    return key, nonce


def is_tidal_api_reachable(hostname: str = "api.tidal.com") -> bool:
    """Using stdlib 'socket' library, test if a few conditions are all
    met: whether the user has a connection to the larger Internet;
    whether the user can resolve the primary URL for this service,
    api.tidal.com, and whether api.tidal.com is responding to requests"""
    try:
        _ = closing(socket.create_connection((hostname, 80)))
    except ConnectionRefusedError:
        logger.critical("It seems that 'api.tidal.com' is unreachable!")
        return False
    except socket.gaierror:
        logger.critical(
            f"tidal-wave is unable to find the IP address of {hostname}: "
            "Please ensure that Internet connectivity is established, "
            "particularly DNS resolution"
        )
        return False
    except OSError as ose:
        if "[Errno 101] Network is unreachable" in ose.msg:
            logger.critical(
                "tidal-wave appears to be unable to reach the Internet. "
                "Please ensure that connectivity to (at least) api.tidal.com "
                "is possible"
            )
            return False
    except Exception as e:
        logger.exception(e)
        return False
    else:
        return True



"""Interact with TIDAL API in order to authenticate tidal-wave as a client."""

from __future__ import annotations

import base64
import json
import logging
import sys
from enum import Enum
from typing import TYPE_CHECKING

import requests
import typer

from .models import BearerAuth, SessionsEndpointResponseJSON
from .oauth import (
    TOKEN_DIR_PATH,
    BearerToken,
    TidalOauth,
    TokenError,
)
from .utils import TIDAL_API_URL

if TYPE_CHECKING:
    from pathlib import Path


COMMON_HEADERS: dict[str, str] = {"Accept-Encoding": "gzip, deflate, br"}
COUNTRY_CODE_PROPER_LENGTH: int = 2

logger = logging.getLogger(__name__)


class AudioFormat(str, Enum):
    """A simple representation of TIDAL's music quality levels."""

    dolby_atmos = "Atmos"
    hi_res = "HiRes"
    lossless = "Lossless"
    high = "High"
    low = "Low"


class LogLevel(str, Enum):
    """A simple representation of logging library's verbosity levels."""

    debug = "DEBUG"  # 10
    info = "INFO"  # 20
    warning = "WARNING"  # 30
    error = "ERROR"  # 40
    critical = "CRITICAL"  # 50


def load_token_from_disk(
    token_path: Path = TOKEN_DIR_PATH / "android-tidal.token",
) -> dict | None:
    """Attempt to read `token_path` from disk and decode its contents as JSON."""
    if not token_path.exists():
        _msg: str = f"FileNotFoundError: {token_path.absolute()}"
        logger.warning(_msg)
        return None
    token_file_contents: str = token_path.read_bytes()
    decoded_token_file_contents: str = base64.b64decode(token_file_contents).decode(
        "utf-8",
    )

    try:
        bearer_token_json: dict = json.loads(decoded_token_file_contents)
    except json.decoder.JSONDecodeError:
        _msg: str = f"File '{token_path.absolute()}' cannot be parsed as JSON"
        logger.warning(_msg)
        return None
    else:
        return bearer_token_json


def validate_token_for_session(
    token: str,
    headers: dict[str, str] = COMMON_HEADERS,
) -> requests.Session | None:
    """Send a GET request to the /sessions endpoint of TIDAL's API.

    If `token` is valid, use the SessionsEndpointResponseJSON object
    that was returned from the API to create a requests.Session object with
    some additional attributes. Otherwise, return None
    """
    auth_headers: dict[str, str] = {**headers, "Authorization": f"Bearer {token}"}
    sess: requests.Session | None = None

    with requests.get(
        url=f"{TIDAL_API_URL}/sessions",
        headers=auth_headers,
        timeout=10,
    ) as r:
        try:
            r.raise_for_status()
        except requests.HTTPError:
            if r.status_code == 401:
                logger.exception("Token is not authorized")
                return sess
            logger.exception("Error occurred when attempting GET request")
            return sess

        serj = SessionsEndpointResponseJSON.from_dict(r.json())
        logger.debug("Adding data from API reponse to session object:")
        logger.debug(serj)

    sess: requests.Session = requests.Session()
    sess.headers = headers
    sess.auth = BearerAuth(token=token)
    sess.user_id = serj.user_id
    sess.session_id = serj.session_id
    sess.client_id = serj.client.id
    sess.client_name = serj.client.name
    if serj.country_code == "US":
        sess.params["countryCode"] = "US"
        sess.params["locale"] = "en_US"
        sess.headers["Accept-Language"] = "en-US"
    elif (len(serj.country_code) == COUNTRY_CODE_PROPER_LENGTH) and (
        serj.country_code.isupper()
    ):
        sess.params["countryCode"] = serj.country_code
    return sess


def login_fire_tv(
    token_path: Path = TOKEN_DIR_PATH / "fire_tv-tidal.token",
) -> requests.Session | None:
    """Load `token_path` from disk, initializing a BearerToken from its contents.

    If successful, return a requests.Session object with
    extra attributes set, particular to the emulated client, Fire TV.
    """
    bearer_token = BearerToken.load(p=token_path)
    if bearer_token is not None:
        bearer_token.save(p=token_path)
    else:
        to = TidalOauth()
        bearer_token = to.authorization_code_flow()
        logger.info("Successfully loaded token from disk.")
        bearer_token.save(p=token_path)

    # check if access needs refreshed
    if bearer_token.is_expired:
        logger.warning("TIDAL access token needs refreshing: Attempting now.")
        try:
            bearer_token.refresh()
        except TokenError as te:
            sys.exit(te.args[0])
        else:
            logger.info("Successfully refreshed TIDAL access token")

    s: requests.Session | None = validate_token_for_session(bearer_token.access_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
    else:
        s.params["deviceType"] = "TV"
        s.headers["User-Agent"] = "TIDAL_ANDROID/2.38.0"
        bearer_token.save()
    return s


def login_android(
    token_path: Path = TOKEN_DIR_PATH / "android-tidal.token",
) -> requests.Session | None:
    """Load `token_path` from disk, initializing a BearerToken from its contents.

    If successful, return a requests.Session object with
    extra attributes set, particular to the emulated client, Android
    phone or tablet.
    """
    _msg: str = f"Loading TIDAL access token from '{token_path.absolute()}'"
    logger.info(_msg)
    _token: dict | None = load_token_from_disk(token_path=token_path)
    access_token: str | None = None if _token is None else _token.get("access_token")
    device_type: str | None = None if _token is None else _token.get("device_type")

    if access_token is None:
        logger.warning("Could not load access token from disk")
        access_token: str = typer.prompt(
            "Enter TIDAL access token from an Android (the part after 'Bearer ')",
        )
        dt_input: str = typer.prompt(
            "Is this from device type: phone, tablet, or other? ",
        )
        if dt_input.lower() == "phone":
            device_type = "PHONE"
        elif dt_input.lower() == "tablet":
            device_type = "TABLET"

    s: requests.Session | None = validate_token_for_session(access_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
        if token_path.exists():
            token_path.unlink()
    else:
        _msg: str = f"Access token is valid: saving to '{token_path.absolute()}'"
        logger.info(_msg)
        if device_type is not None:
            s.params["deviceType"] = device_type

        s.params["platform"] = "ANDROID"
        s.headers["User-Agent"] = "TIDAL_ANDROID/1136 okhttp 4.3.0"
        to_write: dict = {
            "access_token": s.auth.token,
            "session_id": s.session_id,
            "client_id": s.client_id,
            "client_name": s.client_name,
            "country_code": s.params["countryCode"],
            "device_type": device_type,
        }
        _msg: str = f"Writing this bearer token to '{token_path.absolute()}'"
        logger.debug(_msg)
        token_path.write_bytes(base64.b64encode(bytes(json.dumps(to_write), "UTF-8")))
    return s


def login_windows(
    token_path: Path = TOKEN_DIR_PATH / "windows-tidal.token",
) -> requests.Session | None:
    """Load `token_path` from disk, initializing a BearerToken from its contents.

    If successful, return a requests.Session object with
    extra attributes set, particular to the emulated client, a Windows
    laptop or desktop.
    """
    _token: dict | None = load_token_from_disk(token_path=token_path)
    access_token: str | None = None if _token is None else _token.get("access_token")
    if access_token is None:
        access_token: str = typer.prompt(
            "Enter TIDAL API access token (the part after 'Bearer ')",
        )

    s: requests.Session | None = validate_token_for_session(access_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
        if token_path.exists():
            token_path.unlink()
    else:
        _msg: str = f"Writing this access token to '{token_path.absolute()}'"
        logger.debug(_msg)
        s.headers["User-Agent"] = (
            "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) TIDAL/2.36.2 Chrome/116.0.5845.228 Electron/26.6.1 Safari/537.36"
        )
        s.headers["Origin"] = s.headers["Referer"] = "https://desktop.tidal.com/"
        s.params["deviceType"] = "DESKTOP"
        to_write: dict = {
            "access_token": s.auth.token,
            "session_id": s.session_id,
            "client_id": s.client_id,
            "client_name": s.client_name,
            "country_code": s.params["countryCode"],
        }
        token_path.write_bytes(base64.b64encode(bytes(json.dumps(to_write), "UTF-8")))
    return s


def login_macos(
    token_path: Path = TOKEN_DIR_PATH / "mac_os-tidal.token",
) -> requests.Session | None:
    """Load `token_path` from disk, initializing a BearerToken from its contents.

    If successful, return a requests.Session object with
    extra attributes set, particular to the emulated client, a macOS
    laptop or desktop
    """
    _token: dict | None = load_token_from_disk(token_path=token_path)
    access_token: str | None = None if _token is None else _token.get("access_token")
    if access_token is None:
        access_token: str = typer.prompt(
            "Enter TIDAL API access token (the part after 'Bearer ')",
        )

    s: requests.Session | None = validate_token_for_session(access_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
        if token_path.exists():
            token_path.unlink()
    else:
        _msg: str = f"Writing this access token to '{token_path.absolute()}'"
        logger.debug()
        s.headers["User-Agent"] = (
            "TIDALPlayer/3.1.4.209 CFNetwork/1494.0.7 Darwin/23.4.0"
        )
        s.headers["x-tidal-client-version"] = "2024.3.14"
        s.headers["Origin"] = s.headers["Referer"] = "https://desktop.tidal.com/"
        s.params["deviceType"] = "DESKTOP"
        to_write: dict = {
            "access_token": s.auth.token,
            "session_id": s.session_id,
            "client_id": s.client_id,
            "client_name": s.client_name,
            "country_code": s.params["countryCode"],
        }
        token_path.write_bytes(base64.b64encode(bytes(json.dumps(to_write), "UTF-8")))
    return s


def login(
    audio_format: AudioFormat,
) -> tuple[requests.Session | None, AudioFormat | str]:
    """Given an audio_format, either log in "automatically" or using user input.

    The "automatic" method is via the Fire TV OAuth 2.0 flow, and the
    alternative is Android-/Windows-/macOS-gleaned API token; the
    latter to be able to access HiRes fLaC audio.
    Return a tuple of a requests.Session object, if no error, and the
    AudioFormat instance passed in; or (None, "") in the event of error.
    """
    high_quality_formats: set[AudioFormat] = {AudioFormat.hi_res}
    fire_tv_formats: set[AudioFormat] = {
        AudioFormat.dolby_atmos,
        AudioFormat.lossless,
        AudioFormat.high,
        AudioFormat.low,
    }
    if audio_format in fire_tv_formats:
        return (login_fire_tv(), audio_format)

    if audio_format in high_quality_formats:
        # If there's already a token, skip the prompt and input rigmarole
        if (TOKEN_DIR_PATH / "android-tidal.token").exists():
            return (login_android(), audio_format)
        if (TOKEN_DIR_PATH / "windows-tidal.token").exists():
            return (login_windows(), audio_format)
        if (TOKEN_DIR_PATH / "mac_os-tidal.token").exists():
            return (login_macos(), audio_format)

        options: set = {"android", "a", "macos", "m", "windows", "w"}
        _input: str = ""
        while _input not in options:
            _input = typer.prompt(
                "For which of Android [a], macOS [m], or Windows [w] would you like "
                "to provide an API token?",
            ).lower()

        _to_return: tuple[None, str] = (None, "")

        if _input in {"android", "a"}:
            _to_return = (login_android(), audio_format)
        elif _input in {"macos", "m"}:
            _to_return = (login_macos(), audio_format)
        elif _input in {"windows", "w"}:
            _to_return = (login_windows(), audio_format)
        return _to_return

    _msg: str = (
        "Please provide one of the following: "
        f"{', '.join(e.value for e in AudioFormat)}"
    )
    logger.critical()
    return _to_return



"""Retrieve data from TIDAL API and save it locally."""

from __future__ import annotations

import logging
from contextlib import closing
from pathlib import Path

import typer
from cachecontrol import CacheControl
from platformdirs import user_music_path
from typing_extensions import Annotated

from .album import Album
from .artist import Artist
from .login import AudioFormat, LogLevel, login
from .mix import Mix
from .models import (
    TidalAlbum,
    TidalArtist,
    TidalMix,
    TidalPlaylist,
    TidalTrack,
    TidalVideo,
    match_tidal_url,
)
from .playlist import Playlist
from .track import Track
from .utils import is_tidal_api_reachable
from .video import Video

__version__ = "2024.11.1"


# https://typer.tiangolo.com/tutorial/options/version/#fix-with-is_eager
def version_callback(value: bool) -> None:  # noqa: FBT001
    """Pass this function to typer to specify eager option behavior."""
    if value:
        print(f"tidal-wave {__version__}")  # noqa: T201
        raise typer.Exit


app = typer.Typer()
_user_music_path: Path = user_music_path()


@app.command()
def main(
    tidal_url: Annotated[
        str,
        typer.Argument(
            help="The URL to the TIDAL resource that is desired to retrieve.",
        ),
    ],
    audio_format: Annotated[
        AudioFormat,
        typer.Option(case_sensitive=False),
    ] = AudioFormat.lossless.value,
    output_directory: Annotated[
        Path,
        typer.Argument(
            help="The directory under which directory(ies) of files will be written",
        ),
    ] = _user_music_path,
    loglevel: Annotated[
        LogLevel, typer.Option(case_sensitive=False),
    ] = LogLevel.info.value,
    include_eps_singles: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--include-eps-singles",
            help=(
                "No-op unless passing TIDAL artist. Whether to include artist's EPs and"
                " singles with albums"
            ),
        ),
    ] = False,
    no_extra_files: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-extra-files",
            help=(
                "Whether to not even attempt to retrieve artist bio, artist image, "
                "album credits, album review, or playlist m3u8"
            ),
        ),
    ] = False,
    no_flatten: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--no-flatten",
            help=(
                "Whether to treat playlists or mixes as a list of tracks/videos and, as"
                " such, retrieve them independently"
            ),
        ),
    ] = False,
    transparent: Annotated[  # noqa: FBT002
        bool,
        typer.Option(
            "--transparent",
            help="Whether to dump JSON responses from TIDAL API; maximum verbosity",
        ),
    ] = False,
    version: Annotated[  # noqa: ARG001
        bool | None,
        typer.Option("--version", callback=version_callback, is_eager=True),
    ] = None,
):
    """Parse command line arguments and retrieve data from TIDAL."""
    logging.basicConfig(
        format="%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d:%H:%M:%S",
        level=logging.getLevelName(loglevel.value),
    )
    logger = logging.getLogger(__name__)

    tidal_resource: (
        TidalAlbum | TidalMix | TidalPlaylist | TidalTrack | TidalVideo | None
    ) = match_tidal_url(tidal_url)

    if tidal_resource is None:
        _msg: str = (
            f"Cannot parse '{tidal_url}' as a TIDAL album, artist, mix, playlist, "
            "track, or video URL"
        )
        logger.critical(_msg)
        raise typer.Exit(code=1)

    # Check Internet connectivity, and whether api.tidal.com is up
    if not is_tidal_api_reachable():
        user_wishes_to_continue: bool = typer.confirm(
            "\nEven though tidal-wave cannot seem to connect to the Internet, "
            "would you like program execution to continue?",
        )
        if not user_wishes_to_continue:
            raise typer.Exit(code=1)

    s, audio_format = login(audio_format=audio_format)
    if s is None:
        raise typer.Exit(code=1)

    with closing(CacheControl(s)) as session:
        if isinstance(tidal_resource, TidalTrack):
            track = Track(track_id=tidal_resource.tidal_id, transparent=transparent)
            track.get(
                session=session,
                audio_format=audio_format,
                out_dir=output_directory,
                no_extra_files=no_extra_files,
            )

            if loglevel == LogLevel.debug:
                track.dump()
            raise typer.Exit(code=0)
        if isinstance(tidal_resource, TidalAlbum):
            album = Album(album_id=tidal_resource.tidal_id, transparent=transparent)
            album.get(
                session=session,
                audio_format=audio_format,
                out_dir=output_directory,
                no_extra_files=no_extra_files,
            )

            if loglevel == LogLevel.debug:
                album.dump()
            raise typer.Exit(code=0)
        if isinstance(tidal_resource, TidalArtist):
            artist = Artist(artist_id=tidal_resource.tidal_id, transparent=transparent)
            artist.get(
                session=session,
                audio_format=audio_format,
                out_dir=output_directory,
                include_eps_singles=include_eps_singles,
                no_extra_files=no_extra_files,
            )
            raise typer.Exit(code=0)
        if isinstance(tidal_resource, TidalVideo):
            video = Video(video_id=tidal_resource.tidal_id, transparent=transparent)
            video.get(session=session, out_dir=output_directory)

            if loglevel == LogLevel.debug:
                video.dump()
            raise typer.Exit(code=0)
        if isinstance(tidal_resource, TidalPlaylist):
            playlist = Playlist(
                playlist_id=tidal_resource.tidal_id, transparent=transparent,
            )
            if no_flatten:
                playlist.get_elements(
                    session=session,
                    audio_format=audio_format,
                    out_dir=output_directory,
                    no_extra_files=no_extra_files,
                )
            else:
                playlist.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=output_directory,
                    no_extra_files=no_extra_files,
                )

            if loglevel == LogLevel.debug:
                playlist.dump()
            raise typer.Exit(code=0)
        if isinstance(tidal_resource, TidalMix):
            mix = Mix(mix_id=tidal_resource.tidal_id, transparent=transparent)
            if no_flatten:
                mix.get_elements(
                    session=session,
                    audio_format=audio_format,
                    out_dir=output_directory,
                    no_extra_files=no_extra_files,
                )
            else:
                mix.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=output_directory,
                    no_extra_files=no_extra_files,
                )

            if loglevel == LogLevel.debug:
                mix.dump()
            raise typer.Exit(code=0)

        raise NotImplementedError


if __name__ == "__main__":
    app()



"""Represent the DASH manifests retrieved from TIDAL API."""

from __future__ import annotations

import json
import logging
import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from xml.etree import ElementTree

import dataclass_wizard

from .utils import decrypt_manifest_key_id

if TYPE_CHECKING:
    from requests import Session

    from .models import TracksEndpointStreamResponseJSON

logger = logging.getLogger("__name__")


class TidalManifestError(Exception):
    """Exception class to alert issue with parsing of DASH manifest."""


@dataclass
class S:
    d: str
    r: str | None = field(default=None)

    def __post_init__(self) -> None:
        """Set two attributes, d, and r, after dataclass.__init__() executes."""
        self.d: int | None = int(self.d) if self.d is not None else None
        self.r: int | None = int(self.r) if self.r is not None else None


@dataclass(frozen=True)
class SegmentTimeline:
    """Represent a portion of an XML document specifying a DASH manifest."""

    s: tuple[S | None]


@dataclass
class JSONDASHManifest:
    """Represent a JSON document specifying a DASH manifest."""

    mime_type: str | None = field(default=None)
    codecs: str | None = field(default=None)
    encryption_type: str | None = field(default=None)
    key_id: str | None = field(default=None)
    urls: list[str] | None = field(repr=False, default=None)

    def __post_init__(self) -> None:
        """Set two attributes, key and nonce, after dataclass.__init__() executes."""
        self.key: bytes | None = None
        self.nonce: bytes | None = None
        if self.encryption_type == "OLD_AES" and len(self.key_id) > 0:
            logger.debug("Attempting to create decryption key from DASH manifest")
            try:
                self.key, self.nonce = decrypt_manifest_key_id(self.key_id)
            except Exception:
                logger.exception(
                    "An error occurred in the process of decrypting DASH manifest!",
                )


@dataclass
class XMLDASHManifest:
    """Represent an XML document specifying a DASH manifest."""

    mime_type: str | None = field(default=None)
    codecs: str | None = field(default=None)
    content_type: str | None = field(default=None)
    bandwidth: str | None = field(default=None)
    audio_sampling_rate: str | None = field(default=None)
    timescale: str | None = field(default=None)
    initialization: str | None = field(default=None, repr=False)
    media: str | None = field(default=None, repr=False)
    start_number: str | None = field(default=None, repr=False)
    segment_timeline: SegmentTimeline | None = field(default=None, repr=False)

    def __post_init__(self) -> None:
        """Set several instance attributes after __init__() executes."""
        # Initialize key and nonce even though they won't be used in
        # this manifest class
        self.key, self.nonce = None, None

        self.bandwidth: int | None = (
            int(self.bandwidth) if self.bandwidth is not None else None
        )
        self.audio_sampling_rate: int | None = (
            int(self.audio_sampling_rate)
            if self.audio_sampling_rate is not None
            else None
        )
        self.timescale: int | None = (
            int(self.timescale) if self.timescale is not None else None
        )
        self.startNumber: int | None = (
            int(self.start_number) if self.start_number is not None else None
        )

    def build_urls(self, session: Session) -> list[str] | None:
        """Parse the MPEG-DASH manifest into a list of URLs.

        In particular, look for a special value, r, in self.segment_timeline.s.
        If there is no such value, set r=1. In both cases, start substituting
        r into the special substring, '$Number$', in self.initialization.
        Continue incrementing r and substituting until the resulting string
        returns a 500 error to a HEAD request.
        """
        http_code_500: int = 500
        if len(self.segment_timeline.s) == 0:
            return None

        def sub_number(n: int, p: str = r"\$Number\$", s: str = self.media) -> str:
            return re.sub(p, str(n), s)

        try:
            r: int | None = next(S.r for S in self.segment_timeline.s)
        except StopIteration:
            r = None

        # New path for when r is None; e.g. TIDAL track 96154223
        if r is None:
            urls_list: list[str] = [self.initialization]
            number: int = 1
            while session.head(url=sub_number(number)).status_code != http_code_500:
                urls_list.append(sub_number(number))
                number += 1
            return urls_list

        number_range = range(self.startNumber, r + 1)  # include value of `r`
        urls_list: list[str] = [self.initialization] + [
            sub_number(i) for i in number_range
        ]
        number: int = r + 1
        while session.head(url=sub_number(number)).status_code != http_code_500:
            urls_list.append(sub_number(number))
            number += 1
        return urls_list


Manifest = JSONDASHManifest | XMLDASHManifest


def manifester(tesrj: TracksEndpointStreamResponseJSON) -> Manifest:
    """Attempt to return a Manifest-type object based on the attributes of `tesrj`.

    Will raise TidalManifestError upon error.
    """
    if tesrj.manifest_mime_type == "application/vnd.tidal.bts":
        if tesrj.audio_mode not in {"DOLBY_ATMOS", "SONY_360RA", "STEREO"}:
            _msg: str = (
                "Expected a manifest of Dolby Atmos, MQA, Sony 360 Reality Audio, "
                f"or encrypted-for-Windows-client audio for track {tesrj.track_id}"
            )
            raise TidalManifestError(_msg)

        try:
            manifest: Manifest = dataclass_wizard.fromdict(
                JSONDASHManifest, json.loads(tesrj.manifest_bytes)
            )
        except json.decoder.JSONDecodeError as jde:
            _msg: str = (
                "Cannot parse manifest with type "
                f"'{tesrj.manifest_mime_type}' as JSON"
            )
            raise TidalManifestError(_msg) from jde
        except dataclass_wizard.errors.ParseError as pe:
            raise TidalManifestError from pe

        if manifest.encryption_type == "NONE":
            return manifest

        if manifest.encryption_type == "OLD_AES":
            if (manifest.key is not None) and (manifest.nonce is not None):
                return manifest
            _msg: str = (
                f"Audio data for track {tesrj.track_id}, audio mode "
                f"{tesrj.audio_mode} could not be decrypted"
            )
        else:
            _msg: str = (
                f"Audio data for track {tesrj.track_id}, audio mode "
                f"{tesrj.audio_mode} is incorrigibly encrypted with "
                f"encryption type '{manifest.encryption_type}'"
            )
        raise TidalManifestError(_msg)
    elif tesrj.manifest_mime_type == "application/dash+xml":
        try:
            xml: ElementTree.Element = ElementTree.fromstring(tesrj.manifest_bytes)
        except ElementTree.ParseError as pe:
            _msg: str = f"Expected an XML manifest for track {tesrj.track_id}"
            raise TidalManifestError(_msg) from pe

        ns: str = re.match(r"({.*})", xml.tag).groups()[0]
        st: SegmentTimeline = SegmentTimeline(
            tuple(
                S(**el.attrib) if el is not None else None
                for el in xml.findall(f".//{ns}S")
            )
        )

        return XMLDASHManifest(
            xml.find(f".//{ns}AdaptationSet").get("mimeType"),
            xml.find(f".//{ns}Representation").get("codecs"),
            xml.find(f".//{ns}AdaptationSet").get("contentType"),
            xml.find(f".//{ns}Representation").get("bandwidth"),
            xml.find(f".//{ns}Representation").get("audioSamplingRate"),
            xml.find(f".//{ns}SegmentTemplate").get("timescale"),
            xml.find(f".//{ns}SegmentTemplate").get("initialization"),
            xml.find(f".//{ns}SegmentTemplate").get("media"),
            xml.find(f".//{ns}SegmentTemplate").get("startNumber"),
            st,
        )
    else:
        raise TidalManifestError("Manifest MIME type passed is not recognized.")
        return None



"""Represent a playlist in the reckoning of the TIDAL API."""

from __future__ import annotations

import json
import logging
import math
import shutil
import sys
from dataclasses import dataclass
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING
from uuid import uuid4

import ffmpeg
import mutagen
from requests import HTTPError, Session

from .models import (
    PlaylistsEndpointResponseJSON,
    TracksEndpointResponseJSON,
    VideosEndpointResponseJSON,
)
from .requesting import request_playlists
from .track import Track
from .utils import (
    TIDAL_API_URL,
    download_cover_image,
    replace_illegal_characters,
    temporary_file,
)
from .video import Video

if TYPE_CHECKING:
    from .media import AudioFormat

logger = logging.getLogger("__name__")


@dataclass
class Playlist:
    playlist_id: str  # UUID4
    transparent: bool = False

    def __post_init__(self):
        self.playlist_dir: Path | None = None
        self.playlist_cover_saved: bool = False

    def set_metadata(self, session: Session) -> None:
        """Request from TIDAL API /playlists endpoint."""
        self.metadata: PlaylistsEndpointResponseJSON | None = request_playlists(
            session=session,
            playlist_id=self.playlist_id,
            transparent=self.transparent,
        )

        if self.metadata is None:
            return

        self.name = replace_illegal_characters(self.metadata.title)

    def set_items(self, session: Session):
        """Populate self.items using data from TIDAL API /playlists/items endpoint.

        If 'totalNumberOfItems' field returned
        has value greater than 100, multiple requests of size <= 100 will be
        sent to the endpoint until all items for the playlist are retrieved.
        """
        playlist_items: PlaylistsItemsResponseJSON | None = retrieve_playlist_items(
            session=session, playlist_id=self.playlist_id,
        )
        if playlist_items is None:
            self.items = ()
        else:
            self.items: tuple[PlaylistItem] = tuple(
                filter(None, playlist_items.items),
            )

    def set_playlist_dir(self, out_dir: Path):
        """Populate self.playlist_dir based on self.name, self.playlist_id."""
        playlist_substring: str = f"{self.name} [{self.playlist_id}]"
        self.playlist_dir: Path = out_dir / "Playlists" / playlist_substring
        self.playlist_dir.mkdir(parents=True, exist_ok=True)

    def save_cover_image(self, session: Session, out_dir: Path):
        """Request self.metadata.image and attempts to write it to disk."""
        if self.playlist_dir is None:
            self.set_playlist_dir(out_dir=out_dir)
        self.cover_path: Path = self.playlist_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.square_image,
                output_dir=self.playlist_dir,
                dimension=1080,
            )
        else:
            self.playlist_cover_saved = True

    def save_description(self):
        """Request self.metadata.description and attempts to write it to disk."""
        self.description_path: Path = self.playlist_dir / "PlaylistDescription.txt"
        if (
            self.metadata.description is not None
            and len(self.metadata.description) > 0
            and not self.description_path.exists()
        ):
            self.description_path.write_text(f"{self.metadata.description}\n")

    def get_items(
        self,
        session: Session,
        audio_format: AudioFormat,
        no_extra_files: bool,
    ) -> tuple[Track | Video | None] | None:
        """Using either Track.get() or Video.get(), attempt to request
        the data for each track or video in self.items. If no_extra_files
        is True, do not attempt to retrieve or save any of: playlist
        description text, playlist m3u8 text, playlist cover image."""
        if len(self.items) == 0:
            return None
        tracks_videos: list = [None] * len(self.items)
        for i, item in enumerate(self.items):
            if item is None:
                tracks_videos[i] = None
                continue
            if isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id, transparent=self.transparent)
                track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=self.playlist_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                    origin_jpg=False,
                )
                tracks_videos[i] = track
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id, transparent=self.transparent)
                video.get(
                    session=session,
                    out_dir=self.playlist_dir,
                    metadata=item,
                )
                tracks_videos[i] = video
            else:
                tracks_videos[i] = None
                continue
        self.tracks_videos: tuple[Track | Video | None] = tuple(tracks_videos)
        return tracks_videos

    def flatten_playlist_dir(self):
        """When self.get_items() is called, the tracks and/or videos in
        self.items are downloaded using their self-contained .get() logic;
        this means that they will be downloaded to albums. This function
        "flattens" self.playlist_dir, meaning that it moves all downloaded
        audio and video files to self.playlist_dir, and removes the various
        subdirectories created"""
        files: list[dict[int, str | None]] = [None] * len(self.tracks_videos)
        if len(self.tracks_videos) == 0:
            return None
        subdirs: set[Path] = set()

        for i, tv in enumerate(self.tracks_videos, 1):
            if tv.outfile is None:
                try:
                    _ = tv.album_dir
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
                files[i - 1] = {i: None}
                continue

            _path: Path | None = Path(tv.outfile) if tv is not None else None
            # if the item never got turned into a track or video
            if _path is None:
                files[i - 1] = {i: None}
                continue

            # if the track or video didn't download
            if _path.exists():
                if _path.stat().st_size == 0:
                    files[i - 1] = {i: None}
                    continue
            else:
                files[i - 1] = {i: None}
                continue

            # otherwise, move files and clean up
            if isinstance(tv, Track):
                new_path: Path = self.playlist_dir / f"{i:03d} - {tv.trackname}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
            elif isinstance(tv, Video):
                new_path: Path = self.playlist_dir / f"{i:03d} - {_path.name}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
        self.files: list[dict[int, str | None]] = files

        # Find all subdirectories written to
        for tv in self.tracks_videos:
            if isinstance(tv, Track):
                try:
                    _ = tv.album_dir
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
            elif isinstance(tv, Video):
                subdirs.add(tv.artist_dir)

        # Copy all artist images, artist bio JSON files out
        # of subdirs
        artist_images: set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*.jpg"):
                if p.name == "cover.jpg":
                    continue
                artist_images.add(p)
        for artist_image_path in artist_images:
            if artist_image_path.exists():
                shutil.copyfile(
                    artist_image_path.absolute(),
                    self.playlist_dir / artist_image_path.name,
                )

        artist_bios: set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*bio.json"):
                artist_bios.add(p)
        for artist_bio_path in artist_bios:
            if artist_bio_path.exists():
                shutil.copyfile(
                    artist_bio_path.absolute(),
                    self.playlist_dir / artist_bio_path.name,
                )

        # Remove all subdirs
        for subdir in subdirs:
            if subdir.exists():
                shutil.rmtree(subdir)
        return self.playlist_dir

    def craft_m3u8_text(self):
        """This method creates a file called playlist.m3u8 in self.playlist_dir
        that is a standard M3U. Needs to be called after self.flatten_playlist_dir
        in order to be able to access self.files
        N.b. the already-written file is temporarily copied to a .mp4 version in a
        temporary directory because .m4a files cannot be read with mutagen."""
        m3u_text: str = f"#EXTM3U\n#EXTENC:UTF-8\n#EXTIMG:{str(self.cover_path.absolute())}\n#PLAYLIST:{self.name}\n"

        _msg: str = (
            f"Creating .m3u8 playlist file for Playlist with ID '{self.playlist_id}'"
        )
        logger.info(_msg)
        for d in self.files:
            file: str = next(iter(d.values()))
            if file is None:
                continue
            if file.endswith(".flac"):
                m = mutagen.File(file)
                artist: str = m.get("artist", [""])[0]
                title: str = m.get("title", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".mka"):
                m = mutagen.File(file)
                artist: str = m.get("ARTI", [""])[0]
                title: str = m.get("TITL", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".m4a"):
                # Mutagen cannot read .m4a files, so make a copy with all
                # of the metadata tags as a .mp4 in a temporary directory
                with temporary_file(suffix=".mp4") as tf:
                    ffmpeg.input(file, hide_banner=None, y=None).output(
                        tf.name,
                        acodec="copy",
                        vcodec="copy",
                        loglevel="quiet",
                    ).run()

                    m = mutagen.File(tf.name)
                    artist: str = m.get("\xa9ART", [""])[0]
                    title: str = m.get("\xa9nam", [""])[0]
                    extinf: str = (
                        f"#EXTINF:{math.ceil(m.info.length)},"
                        f"{artist} - {title}\n{file}\n"
                    )
                    m3u_text += extinf
            elif file.endswith(".mp4"):  # video files
                m = mutagen.mp4.MP4(file)
                artist: str = m.get("\xa9ART", [""])[0]
                title: str = m.get("\xa9nam", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
        return m3u_text

    def dumps(self) -> str:
        """This method emulates the stdlib json.dumps(). In particular,
        it returns the JSON-formatted string of the self.files attribute,
        which is an array of objects with one key each: the index in the
        playlist of the track or video, and the absolute file path to the
        track or video"""
        return json.dumps(self.files)

    def dump(self, fp=sys.stdout):
        """This method emulates the stdlib json.dump(). In particular,
        it sends to 'fp' the JSON-formatted string of the self.files attribute,
        which is an array of objects with one key each: the index in the
        playlist of the track or video, and the absolute file path to the
        track or video"""
        json.dump(self.files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ):
        """Execute a number of other methods in a row.

        The methods are:
          - self.set_metadata()
          - self.set_items()
          - self.set_playlist_dir()
          - self.get_items()
          - self.flatten_playlist_dir()
        Then, if no_extra_files is False,
          - self.save_cover_image()
          - self.save_description()
          - self.craft_m3u8_text()
        """
        self.set_metadata(session)

        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        self.set_playlist_dir(out_dir)

        if self.get_items(session, audio_format, no_extra_files) is None:
            _msg: str = f"Could not retrieve playlist with ID '{self.playlist_id}'"
            logger.critical(_msg)
            self.files = {}
            return

        self.flatten_playlist_dir()
        _msg: str = f"Playlist files written to '{self.playlist_dir}'"
        logger.info(_msg)

        if not no_extra_files:
            try:
                self.save_description()
            except Exception:
                logger.exception()
            else:
                _msg: str = (
                    "Playlist description written to "
                    f"{self.playlist_dir / 'PlaylistDescription.txt'}"
                )
                logger.info(_msg)

            self.save_cover_image(session, out_dir)

            try:
                m3u8_text: str = self.craft_m3u8_text()
            except Exception as e:
                _msg: str = (
                    "Unable to create playlist.m3u8 file for "
                    f"playlist with ID '{self.playlist_id}'"
                )
                logger.warning(_msg)
                logger.debug(e)
            else:
                (self.playlist_dir / "playlist.m3u8").write_text(m3u8_text)
                _msg: str = (
                    "Playlist M3U file written to "
                    f"{self.playlist_dir / 'playlist.m3u8'}"
                )
                logger.info(_msg)

    def get_elements(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ) -> None:
        """Execute a number of other methods in a row.

        The main method of this class when no_flatten is True at
        the program top level. The methods executed are:
          - self.set_metadata()
          - self.set_items()
        """
        self.set_metadata(session)

        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        if len(self.items) == 0:
            self.files = {}
            return
        files: list[dict[int, str | None]] = [None] * len(self.items)

        for i, item in enumerate(self.items):
            if item is None:
                files[i] = {i: None}
                continue
            if isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id, transparent=self.transparent)
                track_file: str | None = track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=out_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                    origin_jpg=False,
                )
                files[i] = {i: track_file}
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id, transparent=self.transparent)
                video_file: str | None = video.get(
                    session=session,
                    out_dir=out_dir,
                    metadata=item,
                )
                files[i] = {i: video_file}
            else:
                files[i] = {i: None}
                continue
        self.files: list[dict[int, str | None]] = files


class TidalPlaylistError(Exception):
    """Catch-all custom exception for retrieval of playlist data from TIDAL API."""


def request_playlists_items(
    session: Session,
    playlist_id: str,
    offset: int | None = None,
    transparent: bool = False,
) -> dict | None:
    """Request from TIDAL API /playlists/items endpoint. If requests.HTTPError
    arises, warning is logged; upon this or any other exception, None is returned.
    If no exception arises from 'session'.get(), the requests.Response.json()
    object is returned. If transparent is True, all JSON responses from the API
    are written to disk"""
    url: str = f"{TIDAL_API_URL}/playlists/{playlist_id}/items"
    kwargs: dict = {"url": url}
    kwargs["params"] = (
        {"limit": 100} if offset is None else {"limit": 100, "offset": offset}
    )
    kwargs["headers"] = {"Accept": "application/json"}
    json_name: str = f"playlists-{playlist_id}-items_{uuid4().hex}.json"

    data: dict | None = None
    _msg: str = f"Requesting from TIDAL API: playlists/{playlist_id}/items"
    logger.info(_msg)
    with session.get(**kwargs) as r:
        try:
            r.raise_for_status()
        except HTTPError:
            if r.status_code == 404:
                _msg: str = (
                    "404 Client Error: not found for TIDAL API endpoint "
                    f"playlists/{playlist_id}/items"
                )
                logger.warning(_msg)
            else:
                logger.exception()

        _msg: str = (
            f"{r.status_code} response from TIDAL API to request: "
            f"playlists/{playlist_id}/items"
        )
        if transparent:
            Path(json_name).write_text(
                json.dumps(r.json(), ensure_ascii=True, indent=4, sort_keys=True),
            )
        data = r.json()
        logger.debug(_msg)
        return data


@dataclass(frozen=True)
class PlaylistsItemsResponseJSON:
    """Represent the response from the TIDAL API endpoint /playlists/<ID>/items."""

    limit: int
    offset: int
    total_number_of_items: int
    items: tuple[TracksEndpointResponseJSON | VideosEndpointResponseJSON | None]


def playlists_items_response_json_maker(
    playlists_response: dict[str, int | list[dict]],
) -> PlaylistsItemsResponseJSON | None:
    """This function massages the response from the TIDAL API endpoint
    /playlists/items into a format that PlaylistsItemsResponseJSON.__init__()
    can ingest, and then returns a PlaylistsItemsResponseJSON instance"""
    init_args: dict[str, int | None] = {
        "limit": playlists_response.get("limit"),
        "offset": playlists_response.get("offset"),
        "total_number_of_items": playlists_response.get("totalNumberOfItems"),
    }

    items: tuple[SimpleNamespace] = tuple(
        SimpleNamespace(**d) for d in playlists_response["items"]
    )
    if len(items) == 0:
        return None

    playlist_items: list[
        TracksEndpointResponseJSON | VideosEndpointResponseJSON | None
    ] = [None] * init_args["total_number_of_items"]

    for i, namespace in enumerate(items):
        if namespace.type == "track":
            try:
                playlist_item = TracksEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                _msg: str = (
                    f"TidalPlaylistError: unable to parse playlist item {i} "
                    f"with type '{namespace.type}'"
                )
                logger.warning(_msg)
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        elif namespace.type == "video":
            try:
                playlist_item = VideosEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                _msg: str = (
                    f"TidalPlaylistError: unable to parse playlist item {i} "
                    f"with type '{namespace.type}'"
                )
                logger.warning(_msg)
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        else:
            continue  # value stays None
    init_args["items"] = tuple(playlist_items)

    return PlaylistsItemsResponseJSON(**init_args)


def retrieve_playlist_items(
    session: Session,
    playlist_id: str,
    transparent: bool = False,
) -> PlaylistsItemsResponseJSON | None:
    """The pattern for playlist items retrieval does not follow the
    requesting.request_* functions, hence its implementation here. N.b.
    if the first response from /playlists/items endpoint indicates that
    the playlist contains more than 100 items, multiple requests will be
    sent until all N > 100 items are retrieved."""
    playlists_items_response_json: PlaylistsItemsResponseJSON | None = None
    playlists_response: dict | None = request_playlists_items(
        session=session, playlist_id=playlist_id, transparent=transparent
    )
    if playlists_response is None:
        _msg: str = f"Could not retrieve the items in playlist '{playlist_id}'"
        raise TidalPlaylistError(_msg)

    total_number_of_items: int | None = playlists_response.get("totalNumberOfItems")
    _msg: str = (
        f"Playlist '{playlist_id}' is comprised of {total_number_of_items} items"
    )
    logger.info(_msg)
    if total_number_of_items is None:
        _msg: str = (
            "TIDAL API did not respond with number of items "
            f"in playlist '{playlist_id}'"
        )
        raise TidalPlaylistError(_msg)
    items_to_retrieve: int = total_number_of_items

    all_items_playlist_response: dict = playlists_response

    if total_number_of_items > 100:
        items_list: list[dict] = playlists_response.pop("items")
        offset: int = 100
        while items_to_retrieve > 0:
            pr: dict | None = request_playlists_items(
                session=session, playlist_id=playlist_id, offset=offset
            )

            if (pr is not None) and ((pr_items := pr.get("items")) is not None):
                items_list += pr_items
                offset += 100
                items_to_retrieve -= 100
            else:
                logger.exception(
                    TidalPlaylistError(
                        f"Could not retrieve more than {len(items_list)} "
                        f"elements of playlist '{playlist_id}'. Continuing "
                        "without the remaining "
                        f"{total_number_of_items - len(items_list)}",
                    ),
                )
        all_items_playlist_response["items"] = items_list

    try:
        playlists_items_response_json: PlaylistsItemsResponseJSON | None = (
            playlists_items_response_json_maker(
                playlists_response=all_items_playlist_response,
            )
        )
    except Exception as e:
        logger.exception(TidalPlaylistError(e.args[0]))
    return playlists_items_response_json


# union type for type hinting
PlaylistItem = TracksEndpointResponseJSON | VideosEndpointResponseJSON | None



"""Represent an album in the reckoning of TIDAL API."""

from __future__ import annotations

import contextlib
import json
import logging
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from io import TextIOWrapper
    from pathlib import Path

    from .media import AudioFormat
    from .models import (
        AlbumsCreditsResponseJSON,
        AlbumsEndpointResponseJSON,
        AlbumsItemsResponseJSON,
        AlbumsReviewResponseJSON,
        TracksEndpointResponseJSON,
    )
from requests import RequestException, Session

from .requesting import (
    request_album_review,
    request_albums,
    request_albums_credits,
    request_albums_items,
)
from .track import Track
from .utils import IMAGE_URL, download_cover_image

logger = logging.getLogger("__name__")


@dataclass
class Album:
    """Class to represent an album in the TIDAL API.

    It just needs the album's ID and a Boolean as arguments to the
    constructor; the Boolean specifies whether to write all JSON
    responses from the TIDAL API to disk, for transparency's sake.

    Methods, e.g. `set_tracks()`, are germane to retrieving data from the
    TIDAL API.
    """

    album_id: int
    transparent: bool = False

    def __post_init__(self) -> None:
        """After dataclass's __init__() runs, this method sets two attributes.

        The attributes will be updated during further method calls.
        """
        self.album_dir: Path | None = None
        self.album_cover_saved: bool = False

    def set_tracks(self, session: Session) -> None:
        """Populate the `tracks` attribute of `self`.

        The attribute is populated by requesting from TIDAL API's
        albums/items endpoint and transforming the response JSON.
        """
        album_items: AlbumsItemsResponseJSON | None = request_albums_items(
            session=session,
            album_id=self.album_id,
            transparent=self.transparent,
        )
        _items = album_items.items if album_items is not None else []
        if self.metadata.number_of_tracks > len(_items):
            items_to_retrieve: int = self.metadata.number_of_tracks - len(_items)
            offset: int = 100
            while items_to_retrieve > 0:
                airj: AlbumsItemsResponseJSON | None = request_albums_items(
                    session=session,
                    album_id=self.album_id,
                    transparent=self.transparent,
                    offset=offset,
                )
                if (airj is not None) and (airj.items is not None):
                    _items += airj.items
                    offset += 100
                    items_to_retrieve -= 100
                else:
                    msg: str = (
                        f"Could not retrieve more than {len(_items)} "
                        f"tracks of album '{self.album_id}'. Continuing "
                        "without the remaining "
                        f"{self.metadata.number_of_tracks - len(_items)}"
                    )
                    logger.warning(msg)

        self.tracks: tuple[TracksEndpointResponseJSON] = tuple(
            _item.item for _item in _items
        )

    def set_metadata(self, session: Session) -> None:
        """Set the attribute `metadata` of `self.

        The attribute is populated by requesting from TIDAL API /albums
        endpoint and converting the response JSON.
        """
        self.metadata: AlbumsEndpointResponseJSON = request_albums(
            session=session,
            album_id=self.album_id,
            transparent=self.transparent,
        )

    def set_album_review(self, session: Session) -> None:
        """Request the review text corresponding to self.album_id.

        If an album review exists, it is written to disk as AlbumReview.json
        in self.album_dir.
        """
        self.album_review: AlbumsReviewResponseJSON | None = request_album_review(
            session=session,
            album_id=self.album_id,
            transparent=self.transparent,
        )
        if self.album_review is not None:
            (self.album_dir / "AlbumReview.json").write_text(
                self.album_review.to_json(),
            )

    def set_album_credits(self, session: Session) -> None:
        """Request the album's top-level credits from TIDAL API.

        An album's credits are distinct from each track's credits. The JSON
        data returned from the TIDAL API is converted and stored as
        self.album_credits. If the JSON data returned is not empty,
        then it is written to the file AlbumCredits.json in self.album_dir.
        """
        self.album_credits: AlbumsCreditsResponseJSON | None = request_albums_credits(
            session=session,
            album_id=self.album_id,
            transparent=self.transparent,
        )
        if self.album_credits is not None:
            num_credit: int = len(self.album_credits.credit)
            if num_credit == 0:
                _msg: str = (
                    "No album credits returned from TIDAL API "
                    f"for album {self.album_id}"
                )
                logger.warning(_msg)
            else:
                ac_file: Path = (self.album_dir / "AlbumCredits.json").absolute()
                with ac_file.open("w") as fp:
                    json.dump(obj=self.album_credits.credit, fp=fp)

    def set_album_dir(self, out_dir: Path) -> None:
        """Populate the attribute `album_dir` of `self`.

        Self.album_dir is a sub-subdirectory of out_dir; i.e.
        out_dir/
            <name of the main artist of the album>/
                album_dir/
        """
        artist_substring: str = self.metadata.artist.name.replace("..", "").replace(
            "/",
            "and",
        )
        album_substring: str = (
            f"{self.metadata.name.replace('..', '')} "
            f"[{self.metadata.id}] [{self.metadata.release_date.year}]"
        )
        self.album_dir = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)
        # Create cover_path here, even if the API
        # does not return a cover, to avoid AttributeError later
        self.cover_path: Path = self.album_dir / "cover.jpg"

        if self.metadata.number_of_volumes > 1:
            for v in range(1, self.metadata.number_of_volumes + 1):
                volume_substring: str = f"Volume {v}"
                (out_dir / artist_substring / album_substring / volume_substring).mkdir(
                    parents=True,
                    exist_ok=True,
                )

    def save_cover_image(self, session: Session, out_dir: Path) -> None:
        """Write a file named cover.jpg in self.album_dir.

        This is achieved via the utils.download_cover_image() function.
        If successful, then self.album_cover_saved is set to True.
        """
        if self.album_dir is None:
            self.set_album_dir(out_dir=out_dir)
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.cover,
                output_dir=self.album_dir,
            )
        else:
            self.album_cover_saved = True

    def original_album_cover(self, session: Session) -> None:
        """Write to disk the "original" album cover of a TIDAL album.

        For most albums, TIDAL features the "original", or highest-resolution
        possible image file. This JPEG can be too large to be embedded into FLAC tracks:
        however, it is ideal to have for music archiving purposes.

        This method requests the original cover and overwrites the smaller, 1280x1280
        image that is embedded into each track file. The filename returned by the TIDAL
        API is consistently origin.jpg.

        It is *probably* okay to get this URL in a track.Track method, as HTTP requests
        are cached, so e.g. executing this method for each track in an album won't
        result in many redundant GET requests.
        """
        origin_jpg_url: str = (
            IMAGE_URL % f"{self.metadata.cover.replace('-', '/')}/origin"
        )
        with session.get(url=origin_jpg_url, headers={"Accept": "image/jpeg"}) as resp:
            try:
                resp.raise_for_status()
            except RequestException as re:
                _msg: str = (
                    "Could not retrieve origin.jpg from TIDAL "
                    f"due to error '{re.args[0]}'"
                )
                logger.warning(_msg)
            else:
                (self.album_dir / "cover.jpg").write_bytes(resp.content)

    def get_tracks(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        *,
        no_extra_files: bool,
    ) -> list[str | None]:
        """Call track.Track.get() for each track object in self.tracks.

        The result of each of these calls populates self.track_files.
        """
        track_files: list[str | None] = [None] * self.metadata.number_of_tracks
        for i, t in enumerate(self.tracks):  # type(t) is TracksEndpointResponseJSON
            track: Track = Track(track_id=t.id, transparent=self.transparent)

            track_files_value: str | None = track.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=t,
                album=self.metadata,
                no_extra_files=no_extra_files,
                origin_jpg=False,
            )
            track_files[i] = {track.metadata.track_number: track_files_value}

        self.track_files = track_files

    def dumps(self) -> str:
        """Return a JSON-like string representation of self.track_files."""
        return json.dumps(self.track_files)

    def dump(self, fp: TextIOWrapper = sys.stdout) -> None:
        """Write to `fp` (by default, STDOUT) a JSON-like string of self.track_files."""
        json.dump(self.track_files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: AlbumsEndpointResponseJSON | None = None,
        *,
        no_extra_files: bool = False,
    ) -> None:
        """Execute other methods of self in sequence.

        That is:
            1. set_metadata()
            2. set_tracks()
            3. save_cover_image()
            4. set_album_review()
            5. get_tracks()
        """
        if metadata is None:
            self.set_metadata(session)
        else:
            self.metadata = metadata

        if self.metadata is None:
            self.track_files = {}
            return

        self.set_tracks(session)

        self.set_album_dir(out_dir)

        if self.metadata.cover != "":  # None was sent from the API
            self.save_cover_image(session, out_dir)
        else:
            _msg: str = (
                "No cover image was returned from TIDAL API "
                f"for album {self.album_id}"
            )
            logger.warning(_msg)

        self.get_tracks(session, audio_format, out_dir, no_extra_files=no_extra_files)

        if not no_extra_files:
            self.set_album_review(session)
            self.set_album_credits(session)
            self.original_album_cover(session)
        else:
            with contextlib.suppress(FileNotFoundError):
                self.cover_path.unlink()



"""Represent a track in the reckoning of the TIDAL API."""

from __future__ import annotations

import json
import logging
import re
import shlex
import shutil
import subprocess
import sys
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, TYPE_CHECKING

import ffmpeg
import mutagen
from Crypto.Cipher import AES
from Crypto.Util import Counter
from mutagen.mp4 import MP4Cover
from requests import RequestException

from .dash import (
    JSONDASHManifest,
    Manifest,
    TidalManifestError,
    XMLDASHManifest,
    manifester,
)
from .media import TAG_MAPPING, AudioFormat
from .models import (
    AlbumsEndpointResponseJSON,
    ArtistsBioResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    download_artist_image,
)
from .requesting import (
    fetch_content_length,
    http_request_range_headers,
    request_albums,
    request_artist_bio,
    request_credits,
    request_lyrics,
    request_stream,
    request_tracks,
)
from .utils import IMAGE_URL, download_cover_image, temporary_file

if TYPE_CHECKING:
    from requests import Session

logger = logging.getLogger("__name__")


@dataclass
class Track:
    """Represent an audio track in the reckoning of TIDAL API."""

    track_id: int
    transparent: bool = False

    def __post_init__(self):
        self._has_lyrics: bool | None = None
        self.tags: dict = {}
        self.af_aq: dict[AudioFormat, str] = {
            AudioFormat.dolby_atmos: "LOW",
            AudioFormat.hi_res: "HI_RES",
            AudioFormat.lossless: "LOSSLESS",
            AudioFormat.high: "HIGH",
            AudioFormat.low: "LOW",
        }

    def set_metadata(self, session: Session):
        """Populate self.metadata after executing request_tracks().

        If an error occurs, self.metadata is set to None. Otherwise,
        self.metadata is set to the response of request_tracks(), a
        TracksEndpointResponseJSON instance.
        """
        self.metadata: TracksEndpointResponseJSON | None = request_tracks(
            session=session,
            track_id=self.track_id,
            transparent=self.transparent,
        )

    def set_album(self, session: Session):
        """Populate self.album by querying TIDAL API.

        This method executes request_albums, passing in 'session' and
        self.metadata.album.id. If an error occurs, self.album is set to None.
        Otherwise, self.album is set to the response of request_albums(),
        an AlbumsEndpointResponseJSON instance.
        """
        self.album: AlbumsEndpointResponseJSON | None = request_albums(
            session=session,
            album_id=self.metadata.album.id,
            transparent=self.transparent,
        )

    def set_credits(self, session: Session):
        """Execute request_credits, using the output to populate self.credits.

        If an error occurs, self.credits is set to None. Otherwise, self.credits
        is a TracksCreditsResponseJSON instance.
        """
        self.credits: TracksCreditsResponseJSON | None = request_credits(
            session=session,
            track_id=self.track_id,
            transparent=self.transparent,
        )

    def get_lyrics(self, session: Session) -> str:
        """Execute request_lyrics, using the output to populate self.credits.

        If an error occurs, self.credits is set to None. Otherwise, self.credits
        is a TracksLyricsResponseJSON instance.
        """
        if self._has_lyrics is None:
            self.lyrics: TracksLyricsResponseJSON | None = request_lyrics(
                session=session,
                track_id=self.track_id,
                transparent=self.transparent,
            )
            if self.lyrics is None:
                self._has_lyrics = False
            else:
                self._has_lyrics = True
        else:
            return self.lyrics

    def set_stream(self, session: Session, audio_format: AudioFormat):
        """Populate self.stream.

        The value populated is either None (in the event of request error),
        or TracksEndpointStreamResponseJSON.
        """
        aq: str | None = self.af_aq.get(audio_format)
        self.stream: TracksEndpointStreamResponseJSON | None = request_stream(
            session,
            self.track_id,
            aq,
            transparent=self.transparent,
        )

    def set_manifest(self):
        """Populate self.manifest and self.codec."""
        try:
            self.manifest: Manifest = manifester(self.stream)
        except TidalManifestError as tme:
            logger.critical(tme.args[0])
            self.manifest = None
            self.codec = None
            return

        # https://dashif.org/codecs/audio/
        if self.manifest.codecs == "flac":
            self.codec = "flac"
        elif self.manifest.codecs in {
            "mp4a.40.5",  # HE-AAC
            "mp4a.40.29",  # HE-AAC v2
            "mp4a.40.2",  # AAC-LC
            "eac3",  # Enhanced AC-3
        }:
            self.codec = "m4a"
        elif self.manifest.codecs == "mp4a.40.34":  # MP3
            self.codec = "mp3"

    def set_album_dir(self, out_dir: Path):
        """Populate self.album_dir, based on self.album and out_dir.

        In particular, self.album_dir is a subdirectory of out_dir
        based on the name of the album's artist.
        """
        artist_substring: str = self.album.artist.name.replace("..", "").replace(
            "/",
            "and",
        )
        album_substring: str = (
            f"{self.album.name} [{self.album.id}] [{self.album.release_date.year}]"
        )
        self.album_dir: Path = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)

        # Create cover_path here, even if the API
        # does not return a cover, to avoid AttributeError later
        self.cover_path: Path = self.album_dir / "cover.jpg"

        if self.album.number_of_volumes > 1:
            volume_substring: str = f"Volume {self.metadata.volume_number}"
            (self.album_dir / volume_substring).mkdir(parents=True, exist_ok=True)

    def set_filename(self, audio_format: AudioFormat):
        """Populate self.filename, which is based on self.metadata, audio_format.

        Additionally, if the available codecs in self.manifest don't match
        audio_format, warnings are logged.
        """
        _track_part: str = f"{self.metadata.track_number:02d} - {self.metadata.name}"
        if audio_format == AudioFormat.low:
            track_substring: str = f"{_track_part} [L]"
        elif audio_format == AudioFormat.high:
            track_substring: str = f"{_track_part} [H]"
        elif audio_format == AudioFormat.lossless:
            track_substring: str = f"{_track_part} [CD]"
        elif audio_format == AudioFormat.hi_res:
            track_substring: str = f"{_track_part} [HiRes]"
        elif audio_format == AudioFormat.dolby_atmos:
            track_substring: str = f"{_track_part} [A]"
        else:
            track_substring: str = _track_part

        # Check for MQA masquerading as HiRes here
        if audio_format == AudioFormat.hi_res:
            if (self.stream.bit_depth == 16) and (self.stream.sample_rate == 44_100):
                logger.warning(
                    "Even though HiRes audio format was requested, and TIDAL responded "
                    " to that request without error, this track is only available in "
                    "lossless format; i.e. 16-bit 44.1 kHz quality. Downloading of "
                    "track will continue, but it will be marked as Lossless ([CD]).",
                )
                self.filename: str | None = f"{_track_part} [CD].{self.codec}"
            else:
                self.filename: str | None = f"{track_substring}.{self.codec}"
        else:
            self.filename: str | None = f"{track_substring}.{self.codec}"

        # for use in playlist file ordering
        if self.filename is None:
            self.trackname: str | None = None
        else:
            self.trackname: str = re.match(
                r"(?:\d{2,3} - )(.+?$)",
                self.filename,
            ).groups()[0]

    def set_outfile(self) -> Path | None:
        """Populate self.outfile.

        Use self.album_dir and self.metadata and self.filename
        to craft the pathlib.Path object, self.outfile, that is a
        reference to where the track will be written on disk.
        """
        if self.album.number_of_volumes > 1:
            self.outfile: Path = (
                self.album_dir / f"Volume {self.metadata.volume_number}" / self.filename
            )
            self.absolute_outfile = str(self.outfile.absolute())
        else:
            self.outfile: Path = self.album_dir / self.filename
            self.absolute_outfile = str(self.outfile.absolute())

        # Keep absolute_outfile under 260-character limit
        if len(self.absolute_outfile) >= 260:
            outfile: str = f"{self.metadata.track_number:02d}.{self.codec}"
            if len(outfile) >= 260:
                _msg: str = (
                    f"Unable to fit title of track {self.track_id} "
                    f"into a valid filename (even '{outfile}'), as "
                    "total character count exceeds Windows' 260-character "
                    "limit. Downloading will not continue",
                )
                logger.warning(_msg)
                return

            self.outfile: Path = Path(self.absolute_outfile).parent / outfile
            self.absolute_outfile: str = str(self.outfile.absolute())
            self.trackname: str = outfile

        if (self.outfile.exists()) and (self.outfile.stat().st_size > 0):
            _msg: str = (
                f"Track {self.absolute_outfile} already exists "
                "and therefore will not be overwritten"
            )
            logger.info(_msg)
            return None
        return self.outfile

    def save_artist_image(self, session: Session):
        """Write a JPEG with the name of all self.metadata.artists to self.album_dir."""
        for a in self.metadata.artists:
            track_artist_image: Path = (
                self.album_dir / f"{a.name.replace('..', '').replace('/', 'and')}.jpg"
            )
            if not track_artist_image.exists():
                download_artist_image(session, a, self.album_dir, dimension=750)

    def save_artist_bio(self, session: Session):
        """Write a JSON file to self.album_dir.

        The JSON file written contains the name of each of
        self.metadata.artists to self.album_dir.
        """
        for a in self.metadata.artists:
            track_artist_bio_json: Path = self.album_dir / f"{a.name}-bio.json"
            if not track_artist_bio_json.exists():
                artist_bio: ArtistsBioResponseJSON | None = request_artist_bio(
                    session=session,
                    artist_id=a.id,
                    transparent=self.transparent,
                )
                if artist_bio is not None:
                    _msg: str = (
                        f"Writing artist bio for artist {a.id} to "
                        f"'{track_artist_bio_json.absolute()}"
                    )
                    logger.info(_msg)
                    track_artist_bio_json.write_text(artist_bio.to_json())

    def save_album_cover(self, session: Session):
        """Save cover.jpg to self.album_dir.

        The bytes for cover.jpg come from self.album.cover.
        """
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.album.cover,
                output_dir=self.album_dir,
            )

    def original_album_cover(self, session: Session):
        """Request 'origin.jpg' from TIDAL API for the album of self.track.

        For most albums, TIDAL features the "original" album cover, in the highest
        resolution possible. This JPEG can be too large to be embedded into FLAC tracks,
        however it is ideal to have for music archiving etc. purposes. This method
        requests the original cover and overwrites the smaller, 1280x1280 image used to
        embed into the track file. The filename on the API side is origin.jpg. It is
        *probably* okay to get this URL as a track.Track method, as HTTP requests are
        cached, so e.g. executing this method for each track in an album won't result
        in many redundant GET requests.
        """
        origin_jpg_url: str = IMAGE_URL % f"{self.album.cover.replace('-', '/')}/origin"
        with session.get(url=origin_jpg_url, headers={"Accept": "image/jpeg"}) as resp:
            try:
                resp.raise_for_status()
            except RequestException as re:
                _msg: str = (
                    "Could not retrieve origin.jpg from TIDAL "
                    f"due to error '{re.args[0]}'"
                )
                logger.warning(_msg)
            else:
                (self.album_dir / "cover.jpg").write_bytes(resp.content)

    def set_urls(self, session: Session):
        """Set self.urls based on self.manifest."""
        if isinstance(self.manifest, JSONDASHManifest):
            self.urls: str | None = self.manifest.urls
        elif isinstance(self.manifest, XMLDASHManifest):
            self.urls: str | None = self.manifest.build_urls(session=session)
        self.download_headers: dict[str, str] = {"Accept": self.manifest.mime_type}
        if session.session_id is not None:
            self.download_headers["sessionId"] = session.session_id
        self.download_params = {k: None for k in session.params}

    def download_url(self, session: Session, out_dir: Path) -> Path | None:
        """Download self.urls[0], when the track manifest contains one URL.

        It relies on byte range headers to incrementally get all content from a URL.
        """
        _msg: str = f"Writing track {self.track_id} to '{self.absolute_outfile}'"
        logger.info(_msg)
        # Implement HTTP range requests here to mimic official clients
        range_size: int = 1024 * 1024  # 1 MiB
        content_length: int = fetch_content_length(session=session, url=self.urls[0])
        if content_length == 0:
            return None

        range_headers: Iterable[str] = http_request_range_headers(
            content_length=content_length,
            range_size=range_size,
            return_tuple=False,
        )

        with temporary_file(suffix=".mp4") as ntf:
            for rh in range_headers:
                with session.get(
                    self.urls[0],
                    params=self.download_params,
                    headers={"Range": rh},
                ) as rr:
                    if not rr.ok:
                        _msg: str = f"Could not download {self}"
                        logger.warning(_msg)
                        return None
                    ntf.write(rr.content)
                    _msg: str = f"Wrote {rh} of track {self.track_id} to '{ntf.name}'"
                    logger.debug(_msg)
            ntf.seek(0)
            _msg: str = f"Finished writing track {self.track_id} to '{ntf.name}'"
            logger.debug(_msg)

            if (self.manifest.key is not None) and (self.manifest.nonce is not None):
                _msg: str = (
                    f"Audio data for track {self.track_id} is encrypted. "
                    "Attempting to decrypt now"
                )
                logger.info(_msg)
                counter: dict[str, int | bytes | bool] = Counter.new(
                    64,
                    prefix=self.manifest.nonce,
                    initial_value=0,
                )
                decryptor = AES.new(self.manifest.key, AES.MODE_CTR, counter=counter)

                # decrypt and write to new temporary file
                with temporary_file(suffix=".mp4") as f_decrypted:
                    # I hope that it doesn't come back to bite me, reading the bytes
                    # from a file with an open 'wb' file descriptor context manager
                    audio_bytes: bytes = decryptor.decrypt(Path(ntf.name).read_bytes())
                    f_decrypted.write(audio_bytes)
                    _msg: str = (
                        f"Audio data for track {self.track_id} has been decrypted."
                    )
                    logger.info(_msg)

                    if self.codec == "flac":
                        # Have to use FFMPEG to re-mux the audio bytes, otherwise
                        # mutagen chokes on NoFlacHeaderError
                        _msg: str = (
                            f"Using FFmpeg to remux '{f_decrypted.name}', "
                            f"writing to '{self.absolute_outfile}'"
                        )
                        logger.debug(_msg)
                        ffmpeg.input(f_decrypted.name, hide_banner=None, y=None).output(
                            self.absolute_outfile,
                            acodec="copy",
                            loglevel="quiet",
                        ).run()
                    elif self.codec == "m4a":
                        shutil.copyfile(f_decrypted.name, self.outfile)

                    _msg: str = (
                        f"Track {self.track_id} written to '{self.absolute_outfile}'"
                    )
                    logger.info(_msg)
                    return self.outfile
            else:
                if self.codec == "flac":
                    # Have to use FFMPEG to re-mux the audio bytes, otherwise
                    # mutagen chokes on NoFlacHeaderError
                    _msg: str = (
                        f"Using FFmpeg to remux '{ntf.name}', "
                        f"writing to '{self.absolute_outfile}'"
                    )
                    logger.debug(_msg)
                    ffmpeg.input(ntf.name, hide_banner=None, y=None).output(
                        self.absolute_outfile,
                        acodec="copy",
                        loglevel="quiet",
                    ).run()
                elif self.codec == "m4a":
                    shutil.copyfile(ntf.name, self.outfile)
                _msg: str = (
                    f"Track {self.track_id} written to '{self.outfile.absolute()}'"
                )
                logger.info(_msg)
                return self.outfile

    def download_urls(self, session: Session) -> Path | None:
        """Write the contents from self.urls to a temporary directory.

        Then, uses FFmpeg to re-mux the data to self.outfile.
        """
        _msg: str = f"Writing track {self.track_id} to '{self.absolute_outfile}'"
        logger.info(_msg)

        with temporary_file(suffix=".mp4") as ntf:
            for i, u in enumerate(self.urls, 1):
                _url: str = u.split("?")[0]
                _msg: str = (
                    f"Requesting part {i} of track {self.track_id} "
                    f"from '{_url}', writing to '{ntf.name}'"
                )
                logger.debug(_msg)
                with session.get(
                    url=u, headers=self.download_headers, params=self.download_params
                ) as resp:
                    if not resp.ok:
                        _msg: str = f"Could not download {self}"
                        logger.warning(_msg)
                        return None
                    ntf.write(resp.content)
            ntf.seek(0)

            if (self.manifest.key is not None) and (self.manifest.nonce is not None):
                _msg: str = (
                    f"Audio data for track {self.track_id} is encrypted. "
                    "Attempting to decrypt now"
                )
                logger.info(_msg)
                counter: dict[str, int | bytes | bool] = Counter.new(
                    64,
                    prefix=self.manifest.nonce,
                    initial_value=0,
                )
                decryptor = AES.new(self.manifest.key, AES.MODE_CTR, counter=counter)
                # decrypt and write to new temporary file
                with temporary_file(suffix=".mp4") as f_decrypted:
                    # I hope that it doesn't come back to bite me, reading the bytes
                    # from a file with an open 'wb' file descriptor context manager
                    audio_bytes: bytes = decryptor.decrypt(Path(ntf.name).read_bytes())
                    f_decrypted.write(audio_bytes)
                    _msg: str = (
                        f"Audio data for track {self.track_id} has been decrypted."
                    )
                    logger.info(_msg)

                    if self.codec == "flac":
                        # Have to use FFMPEG to re-mux the audio bytes, otherwise
                        # mutagen chokes on NoFlacHeaderError
                        _msg: str = (
                            f"Using FFmpeg to remux '{f_decrypted.name}', "
                            f"writing to '{self.absolute_outfile}'"
                        )
                        logger.debug(_msg)
                        ffmpeg.input(f_decrypted.name, hide_banner=None, y=None).output(
                            self.absolute_outfile,
                            acodec="copy",
                            loglevel="quiet",
                        ).run()
                    elif self.codec == "m4a":
                        shutil.copyfile(f_decrypted.name, self.outfile)

                    _msg: str = (
                        f"Track {self.track_id} written to '{self.absolute_outfile}'"
                    )
                    logger.info(_msg)
                    return self.outfile
            else:
                if self.codec == "flac":
                    # Have to use FFmpeg to re-mux the audio bytes, otherwise
                    # mutagen chokes on NoFlacHeaderError
                    _msg: str = (
                        f"Using FFmpeg to remux '{ntf.name}', "
                        f"writing to '{self.absolute_outfile}'"
                    )
                    logger.info(_msg)
                    ffmpeg.input(ntf.name, hide_banner=None, y=None).output(
                        self.absolute_outfile, acodec="copy", loglevel="quiet"
                    ).run()
                elif self.codec == "m4a":
                    shutil.copyfile(ntf.name, self.outfile)

                _msg: str = (
                    f"Track {self.track_id} written to '{self.absolute_outfile}'"
                )
                logger.info(_msg)
                return self.outfile

    def download(self, session: Session, out_dir: Path) -> Path | None:
        """Send a GET request for the data in self.urls and write to self.outfile."""
        if len(self.urls) == 1:
            outfile: Path | None = self.download_url(
                session=session,
                out_dir=out_dir,
            )
        else:
            outfile: Path | None = self.download_urls(session=session)

        return outfile

    def craft_tags(self):
        """Populate the attribute self.tags.

        Using the TAG_MAPPING dictionary, add the correct values
        of various metadata tags to the self.tags dict.
        E.g. for .flac files, the album's artist is 'ALBUMARTIST',
        but for .m4a files, the album's artist is 'aART'.
        """
        tags: dict[str, str | float | list[str] | int] = {}
        if self.codec == "flac":
            tag_map = {k: v["flac"] for k, v in TAG_MAPPING.items()}
        elif self.codec == "m4a":
            tag_map = {k: v["m4a"] for k, v in TAG_MAPPING.items()}

        tags[tag_map["album"]] = self.album.title
        tags[tag_map["album_artist"]] = ";".join(a.name for a in self.album.artists)
        tags[tag_map["album_peak_amplitude"]] = (
            f"{self.stream.album_peak_amplitude}"
            if self.stream.album_peak_amplitude is not None
            else None
        )
        tags[tag_map["album_replay_gain"]] = (
            f"{self.stream.album_replay_gain} dB"
            if self.stream.album_replay_gain is not None
            else None
        )
        tags[tag_map["artist"]] = ";".join(a.name for a in self.metadata.artists)
        tags[tag_map["artists"]] = [a.name for a in self.metadata.artists]
        tags[tag_map["barcode"]] = self.album.upc
        tags[tag_map["copyright"]] = self.metadata.copyright
        tags[tag_map["date"]] = str(self.album.release_date)
        tags[tag_map["isrc"]] = self.metadata.isrc
        tags[tag_map["media"]] = "Digital Media"
        tags[tag_map["title"]] = self.metadata.title
        tags[tag_map["track_peak_amplitude"]] = f"{self.metadata.peak}"
        tags[tag_map["track_peak_amplitude"]] = (
            f"{self.metadata.peak}" if self.metadata.peak is not None else None
        )
        tags[tag_map["track_replay_gain"]] = (
            f"{self.metadata.replay_gain} dB"
            if self.metadata.replay_gain is not None
            else None
        )
        # credits
        for tag in ("composer", "engineer", "lyricist", "mixer", "producer", "remixer"):
            try:
                _credits_tag = ";".join(getattr(self.credits, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map[tag]] = _credits_tag
        # lyrics
        try:
            _lyrics = self.lyrics.subtitles
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["lyrics"]] = _lyrics

        if self.codec == "flac":
            tags[tag_map["comment"]] = self.metadata.url
            # track and disk
            tags["DISCTOTAL"] = f"{self.album.number_of_volumes}"
            tags["DISC"] = f"{self.metadata.volume_number}"
            tags["TRACKTOTAL"] = f"{self.album.number_of_tracks}"
            tags["TRACKNUMBER"] = f"{self.metadata.track_number}"
            # instrument-specific
            #     piano
            try:
                piano_credits: str | None = [
                    f"{{{pc}}} (piano)" for pc in self.credits.piano
                ]
            except (TypeError, AttributeError):  # NoneType problems
                pass
            else:
                tags["PERFORMER"] = piano_credits

        elif self.codec == "m4a":
            tags["\xa9url"] = self.metadata.url
            # Whether explicit field, 'rtng', does not have a FLAC counterpart
            if self.metadata.explicit is None:
                tags["rtng"] = (0,)
            elif not self.metadata.explicit:
                tags["rtng"] = (1,)
            elif self.metadata.explicit:
                tags["rtng"] = (2,)

            tags["pcst"] = 0  # I.e. False because we are not working with podcasts here
            tags["stik"] = (1,)  # Music (https://exiftool.org/TagNames/QuickTime.html)

            # Have to convert to bytes the values of the tags starting with '----'
            for k, v in tags.copy().items():
                if k.startswith("----"):
                    if isinstance(v, str):
                        tags[k] = v.encode("UTF-8")
                    elif isinstance(v, list):
                        tags[k] = [s.encode("UTF-8") for s in v]

            tags["trkn"] = [(self.metadata.track_number, self.album.number_of_tracks)]
            tags["disk"] = [(self.metadata.volume_number, self.album.number_of_volumes)]

        self.tags: dict = {k: v for k, v in tags.items() if v is not None}

    def set_mutagen(self):
        """Create self.mutagen, mutagen.File object, pointing to self.outfile."""
        self.mutagen = mutagen.File(self.outfile)
        self.mutagen.clear()
        self.mutagen.save()

    def set_cover_image_tag(self, dimension: int = 1280):
        """Populate the metadata tag corresponding to the album image.

        This is the metadata tag 'covr' in the case of .m4a files,
        and adds a mutagen.flac.Picture() tag to self.mutagen in the case of
        .flac files. It has been split out from self.set_tags so that it
        can be executed BEFORE self.remux(): otherwise, the .m4a
        metadata tags starting with '----com.apple.iTunes:' are lost.
        """
        if self.codec == "flac":
            p = mutagen.flac.Picture()
            p.type = mutagen.id3.PictureType.COVER_FRONT
            p.desc = "Album Cover"
            p.width = p.height = dimension
            p.mime = "image/jpeg"
            p.data = self.cover_path.read_bytes()
            self.mutagen.add_picture(p)
        elif self.codec == "m4a":
            self.mutagen["covr"] = [
                MP4Cover(self.cover_path.read_bytes(), imageformat=MP4Cover.FORMAT_JPEG)
            ]

        self.mutagen.save()

    def set_tags(self):
        """Add self.tags to self.mutagen, and save the changes to disk."""
        self.mutagen.update(**self.tags)
        self.mutagen.save()

    def remux(self):
        """Execute an FFmpeg command via subprocess.

        It is currently not possible to have mutagen set the order of
        the tracks in the audio file; that is, sometimes the album cover
        "video" track is reported as the 0th track by FFmpeg. Some audio
        players cannot decode this as an audio file, so this method
        runs an FFmpeg command externally to make sure that self.outfile
        has the audio data as the first (and, potentially, only) stream
        according to FFmpeg.
        """
        if self.codec == "flac":
            _cmd: str = f"""ffmpeg -hide_banner -loglevel quiet -y -i "%s"
                -map 0:a:0 -map 0:v:0 -map_metadata 0 -c:a copy -c:v copy
                -metadata:s:v title='Album cover' -metadata:s:v comment='Cover (front)'
                -disposition:v attached_pic "{self.absolute_outfile}" """
            if (
                self.mutagen.pictures is not None
                and isinstance(self.mutagen.pictures, list)
                and len(self.mutagen.pictures) == 1
                and isinstance(self.mutagen.pictures[0], mutagen.flac.Picture)
            ):
                with temporary_file(suffix=".flac") as tf:
                    shutil.move(self.absolute_outfile, tf.name)
                    cmd: str | None = shlex.split(_cmd % tf.name)
                    subprocess.run(cmd, check=True)

        elif self.codec == "m4a":
            _cmd: str = f"""ffmpeg -hide_banner -loglevel quiet -y -i
                "{self.absolute_outfile}" -map 0:a:0 -map 0:v:0 -map_metadata 0 -c:a
                copy -c:v copy -movflags +faststart "%s" """
            if (
                self.mutagen.get("covr") is not None
                and isinstance(self.mutagen["covr"], list)
                and len(self.mutagen["covr"]) == 1
                and isinstance(self.mutagen["covr"][0], MP4Cover)
            ):
                with temporary_file(suffix=".mp4") as tf:
                    cmd: str | None = shlex.split(_cmd % tf.name)
                    subprocess.run(cmd, check=True)
                    shutil.copyfile(tf.name, self.absolute_outfile)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: TracksEndpointResponseJSON | None = None,
        album: AlbumsEndpointResponseJSON | None = None,
        no_extra_files: bool = True,
        origin_jpg: bool = True,
    ) -> str | None:
        """Execute several instance methods in sequence, returning path to audio file.

        This is the main driver method of Track. It executes, in order:
          1) self.set_metadata(session)
          2) self.set_album(session)
          3) self.set_album_dir(out_dir)
          4) self.set_credits(session)
          5) self.set_stream(session, audio_format)
          6) self.save_artist_image(session)
          7) self.save_artist_bio(session)
          8) self.get_lyrics(session)
          9) self.set_urls(session)
          10) self.download(session, out_dir)
          11) self.set_mutagen()
          12) self.set_cover_image_tag()
          13) self.remux()
          14) self.craft_tags()
          15) self.set_tags()
          16) self.original_album_cover(session);

        catching Exceptions and attempting to handle edge cases.
        """
        if metadata is None:
            self.set_metadata(session)
        else:
            self.metadata = metadata

        if self.metadata is None:
            self.outfile = None
            return None

        if (audio_format == AudioFormat.dolby_atmos) and (
            "DOLBY_ATMOS" not in self.metadata.media_metadata.tags
        ):
            _msg: str = (
                "Dolby Atmos audio format was requested, but track "
                f"{self.track_id} is not available in Dolby Atmos "
                "format. Downloading of track will not continue."
            )
            logger.warning(_msg)
            self.outfile = None
            return None

        if album is None:
            self.set_album(session)
        else:
            self.album = album

        if self.album is None:
            self.outfile = None
            return None
        self.set_album_dir(out_dir)

        self.set_credits(session)
        self.set_stream(session, audio_format)
        if self.stream is None:
            self.outfile = None
            return None

        self.set_manifest()
        if self.manifest is None:
            self.outfile = None
            return None

        self.set_filename(audio_format)
        if self.filename is None:
            self.outfile = None
            return None

        outfile: Path | None = self.set_outfile()
        if outfile is None:
            if not no_extra_files:
                with suppress(Exception):
                    self.save_artist_image(session)
                    self.save_artist_bio(session)
            return None

        with suppress(Exception):
            self.get_lyrics(session)

        self.set_urls(session)

        if self.download(session, out_dir) is None:
            return None

        self.set_mutagen()

        if self.album.cover == "":
            _msg: str = (
                f"No cover image was returned from TIDAL API for album {self.album.id}"
            )
            logger.warning(_msg)
        else:
            self.save_album_cover(session)
            if self.cover_path.exists() and self.cover_path.stat().st_size > 0:
                self.set_cover_image_tag()

        self.remux()
        self.craft_tags()
        self.set_tags()

        if not no_extra_files:
            with suppress(Exception):
                self.save_artist_image(session)
                self.save_artist_bio(session)

            if origin_jpg:
                with suppress(Exception):
                    self.original_album_cover(session)
        else:
            with suppress(FileNotFoundError):
                self.cover_path.unlink()

        return self.absolute_outfile

    def dump(self, fp=sys.stdout):
        """Emulate stdlib's json.dump in that a JSON-like object is sent to fp.

        particularly, {self.metadata.track_number: self.absolute_outfile}.
        """
        k: int = int(self.metadata.track_number)
        if (self.outfile is None) or (not isinstance(self.outfile, Path)):
            v: str | None = None
        else:
            v: str | None = self.absolute_outfile
        json.dump({k: v}, fp)

    def dumps(self) -> str | None:
        """Return a str version of self.

        This method emulates stdlib json.dumps(). It creates a
        JSON-formatted string from the dict
        {self.metadata.track_number: self.absolute_outfile}
        """
        k: int = int(self.metadata.track_number)
        if (self.outfile is None) or (not isinstance(self.outfile, Path)):
            v: str | None = None
        else:
            v: str | None = self.absolute_outfile
        json.dumps({k: v})
        return None



"""Create helper functions to request data from various TIDAL API endpoints."""

import json
import logging
from functools import partial
from pathlib import Path
from typing import Callable, Iterable, Iterator, Optional, Tuple, Union
from uuid import uuid4

import backoff
from requests import HTTPError, Response, Session

from .models import (
    AlbumsCreditsResponseJSON,
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
)
from .utils import TIDAL_API_URL

logger: logging.Logger = logging.getLogger(__name__)

ResponseJSON = Union[
    AlbumsCreditsResponseJSON,
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
]


def requester_maker(
    session: Session,
    endpoint: str,
    identifier: int,
    url_end: str = "",
    headers: Optional[dict] = None,
    parameters: Optional[dict] = None,
    subclass: Optional["ResponseJSON"] = None,
    credits_flag: bool = False,
    transparent: bool = False,
    offset: Optional[int] = None,
) -> Callable:
    """This function is a function factory: it crafts nearly identical
    versions of the same logic: send a GET request to a certain endpoint;
    if a requests.HTTPError arises, return None; else, transform the
    JSON response into an instance of a subclass of JSONWizard."""

    def function(s, e, i, u, h, p, sc, cf, t, o):
        url: str = f"{TIDAL_API_URL}/{e}/{i}{u}"
        kwargs: dict = {"url": url}
        if p is not None:
            kwargs["params"] = p if offset is None else p | {"offset": offset}
        elif offset is not None:
            kwargs["params"] = {"offset": offset}

        if h is not None:
            kwargs["headers"] = h

        @backoff.on_predicate(
            backoff.expo,
            predicate=lambda r: r.status_code == 429,
            jitter=backoff.random_jitter,
            max_time=15,
            logger=logger,
        )
        def _get(s: Session, request_kwargs: dict) -> Response:
            """Return a requests.Response object from having passed request_kwargs
            to s.get(), optionally retrying if 429 error occurs."""
            with s.get(**request_kwargs) as r:
                return r

        data: Optional[sc] = None
        logger.info(f"Requesting from TIDAL API: {e}/{i}{u}")
        resp: Response = _get(s=s, request_kwargs=kwargs)

        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint {e}/{i}{u}"
                )
                return None
            if resp.status_code == 401:
                logger.warning(
                    f"401 Client Error: Unauthorized for TIDAL API endpoint {e}/{i}{u}"
                )
                return None
            logger.exception(he)
            return None

        if t:
            json_name: str = (
                f"{e}-{i}-{u.strip('/')}_{uuid4().hex}.json"
                if u != ""
                else f"{e}-{i}_{uuid4().hex}.json"
            )
            Path(json_name).write_text(
                json.dumps(resp.json(), ensure_ascii=True, indent=4, sort_keys=True)
            )

        if cf:
            data = sc.from_dict({"credits": resp.json()})
        else:
            data = sc.from_dict(resp.json())

        return data

    return function(
        s=session,
        e=endpoint,
        i=identifier,
        u=url_end,
        h=headers,
        p=parameters,
        sc=subclass,
        cf=credits_flag,
        t=transparent,
        o=offset,
    )


# Functions will curry all arguments except session and identifier, as those
# don't change during runtime, and session is only available once the __main__
# process (on the happy path) creates a requests.Session object. Identifier
# varies with the media type, etc.
def request_albums(
    session: Session, album_id: int, transparent: bool = False
) -> Optional[AlbumsEndpointResponseJSON]:
    """Send a GET request to the /albums endpoint of the TIDAL API. If an
    Exception occurs, return None. Else, return an instance of
    AlbumsEndpointResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=album_id,
        headers={"Accept": "application/json"},
        subclass=AlbumsEndpointResponseJSON,
        transparent=transparent,
    )


def request_albums_items(
    session: Session,
    album_id: int,
    transparent: bool = False,
    offset: Optional[int] = None,
) -> Optional[AlbumsItemsResponseJSON]:
    """Send a GET request to the /albums/<album ID>/items
    endpoint of the TIDAL API. If an Exception occurs,
    return None. Else, return an instance of
    AlbumsItemsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=album_id,
        headers={"Accept": "application/json"},
        parameters={"limit": 100},
        url_end="/items",
        subclass=AlbumsItemsResponseJSON,
        transparent=transparent,
        offset=offset,
    )


def request_album_review(
    session: Session, album_id: int, transparent: bool = False
) -> Optional[AlbumsReviewResponseJSON]:
    """Send a GET request to the /albums/<album ID>/review
    endpoint of the TIDAL API. If an Exception occurs, return None.
    Else, return an instance of AlbumsReviewResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=album_id,
        headers={"Accept": "application/json"},
        url_end="/review",
        subclass=AlbumsReviewResponseJSON,
        transparent=transparent,
    )


def request_artist_bio(
    session: Session, artist_id: int, transparent: bool = False
) -> Optional[ArtistsBioResponseJSON]:
    """Send a GET request to the /artists/<artist ID>/bio
    endpoint of the TIDAL API. If an Exception occurs, return None.
    Else, return an instance of ArtistsBioResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=artist_id,
        headers={"Accept": "application/json"},
        url_end="/bio",
        subclass=ArtistsBioResponseJSON,
        transparent=transparent,
    )


def request_artists(
    session: Session, artist_id: int, transparent: bool = False
) -> Optional[ArtistsEndpointResponseJSON]:
    """Send a GET request to the /artists endpoint of the TIDAL API.
    If an Exception occurs, return None. Else, return an instance of
    ArtistsEndpointResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=artist_id,
        headers={"Accept": "application/json"},
        subclass=ArtistsEndpointResponseJSON,
        transparent=transparent,
    )


def request_artists_albums(
    session: Session, artist_id: int, transparent: bool = False
) -> Optional[ArtistsAlbumsResponseJSON]:
    """Send a GET request to the /artists/<artist ID>/albums
    endpoint of the TIDAL API. If an Exception occurs, return None.
    Else, return an instance of ArtistsAlbumsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=artist_id,
        headers={"Accept": "application/json"},
        url_end="/albums",
        subclass=ArtistsAlbumsResponseJSON,
        transparent=transparent,
    )


def request_artists_audio_works(
    session: Session, artist_id: int, transparent: bool = False
) -> Optional[ArtistsAlbumsResponseJSON]:
    """Send a GET request to the /artists/<artist ID>/albums
    endpoint of the TIDAL API. If an Exception occurs, return None.
    Else, return an instance of ArtistsAlbumsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=artist_id,
        headers={"Accept": "application/json"},
        parameters={"filter": "EPSANDSINGLES"},
        url_end="/albums",
        subclass=ArtistsAlbumsResponseJSON,
        transparent=transparent,
    )


def request_artists_videos(
    session: Session, artist_id: int, transparent: bool = False
) -> Optional[ArtistsVideosResponseJSON]:
    """Send a GET request to the /artists/<artist ID>/videos
    endpoint of the TIDAL API. If an Exception occurs, return None.
    Else, return an instance of ArtistsVideosResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=artist_id,
        headers={"Accept": "application/json"},
        url_end="/videos",
        subclass=ArtistsVideosResponseJSON,
        transparent=transparent,
    )


def request_tracks(
    session: Session, track_id: int, transparent: bool = False
) -> Optional[TracksEndpointResponseJSON]:
    """Send a GET request to the /tracks endpoint of the TIDAL API.
    If an Exception occurs, return None. Else, return an instance of
    TracksEndpointResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=track_id,
        headers={"Accept": "application/json"},
        subclass=TracksEndpointResponseJSON,
        transparent=transparent,
    )


# This one's special, because its JSON response isn't proper JSON:
# it's just an array of JSON objects, so we have to pass a flag to mark
# that the logic common to the rest of the functions is slightly different here.
def request_credits(
    session: Session, track_id: int, transparent: bool = False
) -> Optional[TracksCreditsResponseJSON]:
    """Send a GET request to the /tracks/<track ID>/credits endpoint of the
    TIDAL API. If an Exception occurs, return None. Else, return an instance
    of TracksCreditsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=track_id,
        headers={"Accept": "application/json"},
        parameters={"includeContributors": True},
        url_end="/credits",
        subclass=TracksCreditsResponseJSON,
        credits_flag=True,
        transparent=transparent,
    )


# This one's special, because its JSON response isn't proper JSON:
# it's just an array of JSON objects, so we have to pass a flag to mark
# that the logic common to the rest of the functions is slightly different here.
def request_albums_credits(
    session: Session, album_id: int, transparent: bool = False
) -> Optional[AlbumsCreditsResponseJSON]:
    """Send a GET request to the /albums/<album ID>/credits endpoint of the
    TIDAL API. If an Exception occurs, return None. Else, return an instance
    of AlbumsCreditsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=album_id,
        headers={"Accept": "application/json"},
        parameters={"includeContributors": True, "limit": 50},
        url_end="/credits",
        subclass=AlbumsCreditsResponseJSON,
        credits_flag=True,
        transparent=transparent,
    )


def request_lyrics(
    session: Session, track_id: int, transparent: bool = False
) -> Optional[TracksLyricsResponseJSON]:
    """Send a GET request to the /tracks/<track ID>/lyrics endpoint of the
    TIDAL API. If an Exception occurs, return None. Else, return an instance
    of TracksLyricsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=track_id,
        headers={"Accept": "application/json"},
        url_end="/lyrics",
        subclass=TracksLyricsResponseJSON,
        transparent=transparent,
    )


# One more layer of currying here, as the parameters argument
# is dependent on a runtime variable.
def request_stream(
    session: Session, track_id: int, audio_quality: str, transparent: bool = False
) -> Optional[TracksEndpointStreamResponseJSON]:
    """Send a GET request to the /tracks/<track ID>/playbackinfopostpaywall
    endpoint of the TIDAL API. If an Exception occurs, return None. Else,
    return an instance of TracksEndpointStreamResponseJSON"""
    func = partial(
        requester_maker,
        session=session,
        endpoint="tracks",
        identifier=track_id,
        headers={"Accept": "application/json"},
        parameters={
            "audioquality": audio_quality,
            "playbackmode": "STREAM",
            "assetpresentation": "FULL",
        },
        url_end="/playbackinfopostpaywall",
        subclass=TracksEndpointStreamResponseJSON,
        transparent=transparent,
    )
    return func()


def request_videos(
    session: Session, video_id: int, transparent: bool = False
) -> Optional[VideosEndpointResponseJSON]:
    """Send a GET request to the /videos endpoint of the TIDAL API.
    If an Exception occurs, return None. Else, return an instance of
    VideosEndpointResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="videos",
        identifier=video_id,
        headers={"Accept": "application/json"},
        subclass=VideosEndpointResponseJSON,
        transparent=transparent,
    )


def request_video_contributors(
    session: Session, video_id: int, transparent: bool = False
) -> Optional[VideosContributorsResponseJSON]:
    """Send a GET request to the /videos/<video ID>/contributors endpoint of the
    TIDAL API. If an Exception occurs, return None. Else, return an instance
    of VideosContributorsResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="videos",
        identifier=video_id,
        headers={"Accept": "application/json"},
        parameters={"limit": 100},
        url_end="/contributors",
        subclass=VideosContributorsResponseJSON,
        transparent=transparent,
    )


def request_video_stream(
    session: Session, video_id: int, video_quality: str, transparent: bool = False
) -> Optional[VideosEndpointStreamResponseJSON]:
    """Send a GET request to the /videos/<video ID>/playbackinfopostpaywall
    endpoint of the TIDAL API. If an Exception occurs, return None. Else,
    return an instance of VideosEndpointStreamResponseJSON"""
    func = partial(
        requester_maker,
        session=session,
        identifier=video_id,
        endpoint="videos",
        headers={"Accept": "application/json"},
        parameters={
            "videoquality": video_quality,
            "playbackmode": "STREAM",
            "assetpresentation": "FULL",
        },
        url_end="/playbackinfopostpaywall",
        subclass=VideosEndpointStreamResponseJSON,
        transparent=transparent,
    )
    return func()


def request_playlists(
    session: Session, playlist_id: int, transparent: bool = False
) -> Optional[PlaylistsEndpointResponseJSON]:
    """Send a GET request to the /playlists endpoint of the TIDAL API.
    If an Exception occurs, return None. Else, return an instance of
    PlaylistsEndpointResponseJSON"""
    return requester_maker(
        session=session,
        endpoint="playlists",
        identifier=playlist_id,
        headers={"Accept": "application/json"},
        subclass=PlaylistsEndpointResponseJSON,
        transparent=transparent,
    )


def get_album_id(session: Session, track_id: int) -> Optional[int]:
    """Given the Tidal ID to a track, query the Tidal API in order to retrieve
    the Tidal ID of the album to which the track belongs"""
    terj: Optional[TracksEndpointResponseJSON] = request_tracks(session, track_id)
    album_id: Optional[int] = None

    try:
        album_id = terj.id
    except AttributeError:
        pass
    finally:
        return album_id


def contiguous_ranges(value: int, range_size: int) -> Iterator[Tuple[int, int]]:
    """This function is a generator: it yields two-tuples of int, with the
    tuples representing the (inclusive) boundaries of ranges of size
    range_size. The final tuple will represent a range <= range_size if
    range_size does not evenly divide value. E.g.
    ```>>> list(ranges(16, 3))
    [(0, 2), (3, 5), (6, 8), (9, 11), (12, 14), (15, 16)]
    ```
    N.b. the first tuple will always have first element 0 and the final tuple
    will always have second element `value`."""
    i: int = 0
    rs: int = range_size - 1
    while i + rs < value:
        t: Tuple[int, int] = (i, i + rs)
        i = t[-1] + 1
        yield t
    else:
        yield (i, value)


def http_request_range_headers(
    content_length: int, range_size: int, return_tuple: bool = True
) -> Iterable[str]:
    """This function creates HTTP request Range headers. Its iterable
    returned is of tuples; each tuple describes the (inclusive) boundaries
    of a bytes range with size range_size. If return_tuple is False, it returns
    a generator of tuples. E.g.
    ```>>> http_request_range_headers(16, 3)
    ('bytes=0-2',
     'bytes=3-5',
     'bytes=6-8',
     'bytes=9-11',
     'bytes=12-14',
     'bytes=15-16')
    ```
    """
    ranges: Iterator[Tuple[int, int]] = contiguous_ranges(content_length, range_size)
    iterable: Iterable = (f"bytes={t[0]}-{t[1]}" for t in ranges)
    if return_tuple:
        return tuple(iterable)
    else:
        return iterable


def fetch_content_length(session: Session, url: str) -> int:
    """Attempt to get the amount of bytes pointed to by `url`. If
    the HEAD request from the requests.Session object, `session`,
    encounters an HTTP request; or if the server does not support
    HTTP range requests; or if the server does not respond with a
    Content-Length header, return 0"""
    session_params: dict = session.params
    # Unset params to avoid 403 response
    _params: dict = {k: None for k in session_params}
    with session.head(url=url, params=_params) as resp:
        if not resp.ok:
            cl: str = "0"
        else:
            cl: str = resp.headers.get("Content-Length", "0")
    return int(cl)



"""Represent an artist in the reckoning of TIDAL API."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from .album import Album
from .requesting import (
    request_artists,
    request_artists_albums,
    request_artists_audio_works,
    request_artists_videos,
)
from .utils import download_cover_image
from .video import Video

if TYPE_CHECKING:
    from pathlib import Path

    from requests import Session

    from .media import AudioFormat
    from .models import (
        ArtistsAlbumsResponseJSON,
        ArtistsEndpointResponseJSON,
        ArtistsVideosResponseJSON,
    )

logger = logging.getLogger("__name__")


@dataclass
class Artist:
    """Class to represent an artist in the TIDAL API.

    It just needs the artist's ID and a Boolean as arguments to the
    constructor; the Boolean specifies whether to write all JSON
    responses from the TIDAL API to disk, for transparency's sake.

    Methods, e.g. `set_audio_works()`, are germane to retrieving data from
    the TIDAL API.
    """

    artist_id: int
    transparent: bool = False

    def set_metadata(self, session: Session) -> None:
        """Request from the TIDAL API endpoint /artists.

        Convert the JSON data returned from the API and store the resulting
        object as self.metadata.
        """
        self.metadata: ArtistsEndpointResponseJSON | None = request_artists(
            session=session,
            artist_id=self.artist_id,
            transparent=self.transparent,
        )

    def save_artist_image(self, session: Session) -> None:
        """Write the bytes of self.metadata.picture to disk.

        Specifically, the file cover.jpg in self.artist_dir.
        """
        artist_image: Path = self.artist_dir / "cover.jpg"
        if (not artist_image.exists()) and (self.metadata.picture is not None):
            download_cover_image(
                session,
                self.metadata.picture,
                self.artist_dir,
                dimension=750,
            )

    def set_albums(self, session: Session) -> None:
        """Populate the attribute `self.albums`.

        The JSON data from the TIDAL API endpoint /artists/albums is
        converted and stored as self.albums.
        """
        self.albums: ArtistsAlbumsResponseJSON | None = request_artists_albums(
            session=session,
            artist_id=self.artist_id,
            transparent=self.transparent,
        )

    def set_audio_works(self, session: Session) -> None:
        """Populate self.albums.

        Request from TIDAL API endpoint /artists/albums?filter=EPSANDSINGLES,
        convert the JSON data returned, and store the result as self.albums.
        """
        self.albums: ArtistsAlbumsResponseJSON | None = request_artists_audio_works(
            session=session,
            artist_id=self.artist_id,
            transparent=self.transparent,
        )

    def set_videos(self, session: Session) -> None:
        """Populate self.videos.

        Request from TIDAL API endpoint /artists/videos, convert the JSON
        data returned, and store the results as self.videos.
        """
        self.videos: ArtistsVideosResponseJSON | None = request_artists_videos(
            session=session,
            artist_id=self.artist_id,
            transparent=self.transparent,
        )

    def set_artist_dir(self, out_dir: Path) -> None:
        """Populate self.artist_dir.

        Set self.artist_dir as the subdirectory of `out_dir` simply named the
        value of `self.name`. N.b., a side effect is that the subdirectory on
        the file system is created if it does not exist.
        """
        self.name: str = self.metadata.name.replace("..", "").replace("/", "and")
        self.artist_dir = out_dir / self.name
        self.artist_dir.mkdir(parents=True, exist_ok=True)

    def get_albums(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        *,
        include_eps_singles: bool,
        no_extra_files: bool,
    ) -> list[str | None]:
        """First, fetch all of the albums for `self.artist_id`.

        Then, each of the albums (and, optionally, EPs and singles) is requested and
        written to subdirectories of out_dir.
        """
        if include_eps_singles:
            self.set_audio_works(session)
            _msg: str = (
                f"Starting attempt to get {self.albums.total_number_of_items} "
                "albums, EPs, and singles for artist with ID "
                f"{self.metadata.id},  '{self.name}'"
            )
        else:
            self.set_albums(session)
            _msg: str = (
                f"Starting attempt to get {self.albums.total_number_of_items} albums "
                f"for artist with ID {self.metadata.id}, '{self.name}'"
            )
        logger.info(_msg)

        for a in self.albums.items:
            album: Album = Album(album_id=a.id)
            album.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=a,
                no_extra_files=no_extra_files,
            )

    def get_videos(
        self,
        session: Session,
        out_dir: Path,
    ) -> list[str | None]:
        """Populate `self.videos` by calling self.set_videos().

        Then, for each video, instantiates a Video object and execute
        the object's .get() method.
        """
        self.set_videos(session)
        _msg: str = (
            f"Starting attempt to get {self.videos.total_number_of_items} videos "
            f"for artist with ID {self.metadata.id}, '{self.name}'"
        )
        logger.info(_msg)
        for v in self.videos.items:
            video: Video = Video(video_id=v.id, transparent=self.transparent)
            video.get(
                session=session,
                out_dir=out_dir,
                metadata=v,
            )

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        *,
        include_eps_singles: bool,
        no_extra_files: bool,
    ) -> None:
        """Execute other methods in sequence.

            1. set_metadata()
            2. set_artist_dir()
            3. get_videos()
            4. get_albums()
        Then, if no_extra_files is False, save_artist_image()
        """
        self.set_metadata(session)
        if self.metadata is None:
            return

        self.set_artist_dir(out_dir)
        self.get_videos(session, out_dir)
        if include_eps_singles:
            self.get_albums(
                session,
                audio_format,
                out_dir,
                include_eps_singles=True,
                no_extra_files=no_extra_files,
            )
        self.get_albums(
            session,
            audio_format,
            out_dir,
            include_eps_singles=False,
            no_extra_files=no_extra_files,
        )

        if not no_extra_files:
            self.save_artist_image(session)



"""Represent JSON responses from TIDAL API as JSONWizard subclasses."""

from __future__ import annotations

import base64
import logging
import re
from dataclasses import dataclass, field
from datetime import date, datetime
from io import BytesIO
from typing import TYPE_CHECKING, Literal

import dataclass_wizard
from requests.auth import AuthBase
from typing_extensions import Annotated

from .utils import replace_illegal_characters

if TYPE_CHECKING:
    from pathlib import Path

    from requests import Request, Session


logger = logging.getLogger(__name__)
IMAGE_URL: str = "https://resources.tidal.com/images/%s.jpg"
AudioModeType = Literal["DOLBY_ATMOS", "SONY_360RA", "STEREO"]
AudioQualityType = Literal[
    "HI_RES",
    "HI_RES_LOSSLESS",
    "LOSSLESS",
    "DOLBY_ATMOS",
    "HIGH",
    "LOW",
]
VideoQualityType = Literal["HIGH", "MEDIUM", "LOW", "AUDIO_ONLY"]


@dataclass
class TracksEndpointStreamResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API tracks/{TRACKID} endpoint.

    In particular, the response to a request specifying streaming. The params
    and headers, if correctly specified, return the manifest of the audio to
    be streamed. The manifest is a base64-encoded XML document or JSON object.
    """

    track_id: int
    audio_mode: AudioModeType
    audio_quality: AudioQualityType
    manifest: str = field(repr=False)
    manifest_mime_type: str = field(repr=False)
    album_replay_gain: float | None = field(repr=False, default=None)
    album_peak_amplitude: float | None = field(repr=False, default=None)
    track_replay_gain: float | None = field(repr=False, default=None)
    track_peak_amplitude: float | None = field(repr=False, default=None)
    bit_depth: int | None = field(default=None)
    sample_rate: int | None = field(default=None)

    def __post_init__(self):
        self.manifest_bytes: bytes = base64.b64decode(self.manifest)


class BearerAuth(AuthBase):
    """A class to be passed to `auth` in a `requests.Session` constructor."""

    def __init__(self, token: str):
        self.token = token

    def __call__(self, r: Request):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


@dataclass(frozen=True)
class Client:
    id: int
    name: str
    authorized_for_offline: bool = field(repr=False)


@dataclass(frozen=True)
class SessionsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /sessions."""

    session_id: str  # UUID4 value, really
    user_id: int
    country_code: str  # 2-digit country code according to some ISO
    channel_id: int
    partner_id: int
    client: Client


@dataclass(frozen=True)
class Artist:
    """A musical artist in the reckoning of the TIDAL API."""

    id: int
    name: str
    type: str
    picture: str | None = field(repr=False, default=None)

    def picture_url(self, dimension: int = 320) -> str | None:
        """Create URL pointing to artist's JPEG data."""
        if self.picture is None:
            return None
        if len(self.picture) != 36 or self.picture.count("-") != 4:
            # Should be a UUID
            return None
        _picture = self.picture.replace("-", "/")
        return IMAGE_URL % f"{_picture}/{dimension}x{dimension}"


@dataclass
class MediaMetadata:
    """Represent the sub-object `mediaMetadata` of /tracks and /albums endpoint responses.

    It represents the quality levels available for the album's songs. These
    quality levels are determined by the client device type, the TIDAL account
    level, the country code (read: licensing), the device's quality settings,
    and, perhaps, the device's network connectivity conditions.
    """

    tags: list[str]  # LOSSLESS, HIRES_LOSSLESS


@dataclass(frozen=True)
class TrackAlbum:
    id: int
    title: str
    cover: str = field(repr=None)


@dataclass
class TracksEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint tracks/{TRACKID}.

    If the params and headers are correctly specified, the API returns metadata
    of the available version of the audio track, including audio quality,
    track title, ISRC, track artists, album, track number, duration, etc.
    """

    id: int = field(repr=False)
    title: str
    duration: int  # seconds
    replay_gain: float = field(repr=False)
    peak: float = field(repr=False)
    track_number: int
    volume_number: int
    version: str | None
    copyright: str = field(repr=False)
    url: str
    isrc: str = field(repr=False)
    explicit: bool
    audio_quality: str = field(repr=False)
    audio_modes: list[str] = field(repr=False)
    media_metadata: MediaMetadata
    artist: Artist
    artists: list[Artist]
    album: TrackAlbum

    def __post_init__(self):
        """Set attribute self.name based on values passed to __init__()."""
        name: str = replace_illegal_characters(self.title)
        if self.version is not None:
            version: str = replace_illegal_characters(self.version)
            self.name: str = f"{name} ({version})"
        else:
            self.name: str = name


@dataclass
class AlbumsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /albums/<ALBUMID>.

    If the params and headers are correctly specified, the response should
    contain metadata about the album; e.g. title, number of tracks, copyright,
    date of release, etc.
    """

    id: int = field(repr=False)
    title: str
    duration: int
    number_of_tracks: int
    number_of_volumes: int = field(repr=False)
    release_date: date
    copyright: str = field(repr=False)
    type: str
    version: str | None
    url: str
    cover: str = field(repr=False)
    explicit: bool
    upc: int | str
    audio_quality: str
    audio_modes: list[str]
    media_metadata: MediaMetadata = field(repr=False)
    artist: Artist = field(repr=False)
    artists: list[Artist]

    def __post_init__(self):
        """Set attributes 'name', 'cover_url' based on values passed to __init__()."""
        self.cover_url: str = IMAGE_URL % f"{self.cover.replace('-', '/')}/1280x1280"
        self.name = replace_illegal_characters(self.title)


@dataclass
class AlbumsCreditsResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /albums/<ID>/credits."""

    credits: list[Credit]

    def get_credit(self, type_: str) -> Credit | None:
        """Given a contributor type (e.g. Producer, Engineer),
        go through the `credits` attribute, returning the `Credit` object
        for the given contributor type if it exists"""
        _credit = None
        try:
            _credit = next(c for c in self.credits if c.type == type_)
        except StopIteration:
            _msg: str = f"There are no credits of type {type_} for this album"
            logger.debug(_msg)
            return _credit
        else:
            return _credit

    def get_contributors(self, type_: str) -> list[str] | None:
        """Given a contributor type (e.g. Producer, Engineer),
        go through the `credits` attribute: for each Credit
        object in `self.credits`, if there is a Credit with
        `type` attribute matching `type_` argument, then return
        the `name` attribute for each Contributor object in
        `Credit.contributors`"""
        _credit: Credit | None = self.get_credit(type_)
        _to_return = None
        if _credit is not None:
            _to_return = [c.name for c in _credit.contributors]
        return _to_return

    def __post_init__(self):
        """Create self.credit, a JSON-format amenable dict."""
        _credit: dict[str, str | list[str]] = {
            c: self.get_contributors(c)
            for c in (
                "Cover Design",
                "Creative Director",
                "Design",
                "Engineer",
                "Group Member",
                "Layout",
                "Mastering",
                "Mixing",
                "Photography",
                "Primary Artist",
                "Producer",
                "Record Label",
            )
        }

        self.credit: dict[str, str | list[str]] = {
            k: v for k, v in _credit.items() if v is not None
        }


@dataclass(frozen=True)
class SubscriptionEndpointResponseJSONSubscription:
    """Represent the response from a TIDAL API endpoint.

    In particular, the 'subscription' object that is returned from the
    endpoint /subscription.
    """

    type: str
    offline_grace_period: int


@dataclass(frozen=True)
class SubscriptionEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from a TIDAL API endpoint.

    In particular, the endpoint /subscription.
    """

    start_date: datetime
    valid_until: datetime
    status: str
    subscription: SubscriptionEndpointResponseJSONSubscription
    highest_sound_quality: str
    premium_access: bool
    can_get_trial: bool
    payment_type: str
    payment_overdue: bool


@dataclass(frozen=True)
class AlbumsItemsResponseJSONItem:
    """A sub-object of the response from the TIDAL API endpoint
    /albums/<ID>/items. It simply denotes the type of item, which is surely
    going to be 'track', and is the same object that is returned from the TIDAL
    API /tracks endpoint."""

    item: TracksEndpointResponseJSON
    type: str  # "track"


@dataclass(frozen=True)
class AlbumsItemsResponseJSON(dataclass_wizard.JSONWizard):
    """This class represents the JSON response from the TIDAL API endpoint
    /albums/<ID>/items. It is a list of TracksEndpointResponseJSON objects,
    with a bit of metadata based on the query parameters (offset and limit;
    i.e. pagination logic)."""

    limit: int = field(repr=None)
    offset: int = field(repr=None)
    total_number_of_items: int
    items: list[AlbumsItemsResponseJSONItem]


@dataclass(frozen=True)
class AlbumsReviewResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /albums/<ID>/review."""

    source: str
    last_updated: Annotated[
        datetime,
        dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z"),
    ]
    text: str = field(repr=None)
    summary: str = field(repr=None)


@dataclass(frozen=True)
class Contributor:
    """The response from the TIDAL API endpoint /tracks/<ID>/credits is
    an array of objects, one of the attributes of which is modeled by
    this class. It is simply the name of a contributor to a track,
    and possibly the numerical TIDAL resource ID of that contributor."""

    name: str
    id: int | None = field(repr=False, default=None)


@dataclass(frozen=True)
class Credit:
    """The response from the TIDAL API endpoint /tracks/<ID>/credits is
    an array of objects modeled by this class. It has an attribute,
    `type`, which is one of the roles a person or entity has in the
    creation of a song/album: Composer, Lyricist, Producer, Mixer,
    Engineer, etc. The `contributors` attribute is an array of Name
    and, optionally, TIDAL resource ID for the role"""

    type: str
    contributors: list[Contributor] = field(repr=False)


@dataclass
class TracksCreditsResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /tracks/<ID>/credits."""

    credits: list[Credit]

    def get_credit(self, type_: str) -> Credit | None:
        """Get the credit for the specified type of contributor.

        Given a contributor type (e.g. Lyricist, Composer),
        go through self.credits, returning the `Credit` object
        for the given contributor type if it exists
        """
        _credit = None
        try:
            _credit = next(c for c in self.credits if c.type == type_)
        except StopIteration:
            _msg: str = f"There are no credits of type {type_} for this track"
            logger.debug(_msg)
            return _credit
        else:
            return _credit

    def get_contributors(self, type_: str) -> tuple[str] | None:
        """Given a contributor type (e.g. Lyricist, Composer),
        go through the `credits` attribute: for each Credit
        object in `self.credits`, if there is a Credit with
        `type` attribute matching `type_` argument, then return
        the `name` attribute for each Contributor object in
        `Credit.contributors`"""
        _credit: Credit | None = self.get_credit(type_)
        _to_return: None = None
        if _credit is not None:
            _to_return = tuple(c.name for c in _credit.contributors)
        return _to_return

    def __post_init__(self):
        """Set instance attributes based on self.credits.

        In particular: composer, engineer, lyricist, mixer, producer, remixer, piano.
        """
        self.composer: tuple[str] | None = self.get_contributors("Composer")
        self.engineer: tuple[str] | None = (
            self.get_contributors("Engineer")
            or self.get_contributors("Mastering Engineer")
            or self.get_contributors("Immersive Mastering Engineer")
        )
        self.lyricist: tuple[str] | None = self.get_contributors("Lyricist")
        self.mixer: tuple[str] | None = (
            self.get_contributors("Mixer")
            or self.get_contributors("Mix Engineer")
            or self.get_contributors("Mixing Engineer")
            or self.get_contributors("Atmos Mixing Engineer")
        )
        self.producer: tuple[str] | None = self.get_contributors("Producer")
        self.remixer: tuple[str] | None = self.get_contributors("Remixer")
        self.piano: tuple[str] | None = self.get_contributors("Piano")


@dataclass(frozen=True)
class TracksLyricsResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /tracks/<ID>/lyrics."""

    track_id: int
    lyrics_provider: str
    provider_commontrack_id: str
    provider_lyrics_id: str
    lyrics: str
    subtitles: str
    is_right_to_left: bool


@dataclass
class ArtistsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /artists."""

    id: int
    name: str
    artist_types: list[str]
    url: str
    picture: str | None = field(repr=False, default=None)

    def picture_url(self, dimension: int = 750) -> str | None:
        """Create URL pointing to artist's JPEG data."""
        _to_return: str | None = None
        if self.picture is None:
            return _to_return
        if len(self.picture) != 36 or self.picture.count("-") != 4:
            # Should be a UUID
            return _to_return
        _picture = self.picture.replace("-", "/")
        _to_return: str = IMAGE_URL % f"{_picture}/{dimension}x{dimension}"
        return _to_return


@dataclass(frozen=True)
class ArtistsAlbumsResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /artists/<ID>/albums."""

    limit: int
    offset: int
    total_number_of_items: int
    items: list[AlbumsEndpointResponseJSON]


@dataclass(frozen=True)
class ArtistsVideosResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /artists/<ID>/videos."""

    limit: int
    offset: int
    total_number_of_items: int
    items: list[VideosEndpointResponseJSON]


@dataclass(frozen=True)
class ArtistsBioResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API endpoint /artists/<ID>/bio."""

    source: str
    last_updated: Annotated[
        datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")
    ]
    text: str = field(repr=None)
    summary: str = field(repr=None)


@dataclass
class VideosEndpointStreamResponseJSON(dataclass_wizard.JSONWizard):
    """Represent the response from the TIDAL API videos/<VIDEO_ID> endpoint.

    In particular, when the stream response was requested. The params and
    headers, if correctly specified, return the manifest of the video to be
    streamed. The manifest is a base64-encoded JSON object containing a .m3u8 URL
    """

    video_id: int
    stream_type: str  # ON_DEMAND
    video_quality: VideoQualityType
    manifest: str = field(repr=False)
    manifest_mime_type: str = field(repr=False)

    def __post_init__(self):
        """Set the attribute self.manifest_bytes based on self.manifest."""
        self.manifest_bytes: bytes = base64.b64decode(self.manifest)


@dataclass
class VideosEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API, videos/<VIDEOID> endpoint.If the params and
    headers are correctly specified, the API returns metadata of the available
    version of the (music) video, including video quality, video title, date,
    video artists, duration, etc."""

    id: int = field(repr=False)
    title: str
    volume_number: int
    track_number: int
    release_date: Annotated[
        datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")
    ]
    duration: int  # seconds
    quality: str
    explicit: bool
    type: str
    artist: Artist
    artists: list[Artist]

    def __post_init__(self):
        """Set the attribute self.name based on self.title."""
        self.name = replace_illegal_characters(self.title)


@dataclass(frozen=True)
class VideoContributor:
    """The response from the TIDAL API endpoint /videos/<ID>/credits has
    an array of objects, each of which is modeled by this class.
    It is simply the name of a contributor to a video, and the role of that
    contributor."""

    name: str
    role: str


@dataclass
class VideosContributorsResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /videos/<ID>/contributors
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: list[VideoContributor]

    def get_role(self, role: str) -> tuple[VideoContributor] | None:
        """Given a contributor role (e.g. Composer, Film Director), go through
        `self.items` object, returning the `VideoContributor` object(s)
        for the given contributor type if there are any"""
        role_contributors = tuple(vc for vc in self.items if vc.role == role)
        try:
            role_contributors[0]
        except IndexError:
            _msg: str = f"There are no credits of type '{role}' for this video"
            logger.debug(_msg)
            return None
        else:
            return role_contributors

    def get_contributors(self, role: str) -> tuple[str] | None:
        """Return a tuple of all contributor names of type 'role'."""
        vcs: tuple[VideoContributor] | None = self.get_role(role)
        _to_return: tuple[str] | None = None
        if vcs is not None:
            _to_return = tuple(vc.name for vc in vcs)
        return _to_return

    def __post_init__(self):
        """Set several instance attributes based on self.get_contributors()."""
        self.associated_performer: tuple[str] | None = self.get_contributors(
            "Associated Performer",
        )
        self.composer: tuple[str] | None = self.get_contributors("Composer")
        self.director: tuple[str] | None = self.get_contributors("Director")
        self.engineer: tuple[str] | None = self.get_contributors("Engineer")
        self.film_director: tuple[str] | None = self.get_contributors(
            "Film Director",
        )
        self.film_producer: tuple[str] | None = self.get_contributors(
            "Film Producer",
        )
        self.location: tuple[str] | None = self.get_contributors("Studio")
        self.lyricist: tuple[str] | None = self.get_contributors("Lyricist")
        self.mastering_engineer: tuple[str] | None = self.get_contributors(
            "Mastering Engineer",
        )
        self.mixing_engineer: tuple[str] | None = self.get_contributors(
            "Mixing Engineer",
        )
        self.music_publisher: tuple[str] | None = self.get_contributors(
            "Music Publisher",
        )
        self.producer: tuple[str] | None = self.get_contributors("Producer")
        self.video_director: tuple[str] | None = self.get_contributors(
            "Video Director",
        )
        self.video_producer: tuple[str] | None = self.get_contributors(
            "Video Producer",
        )
        self.vocal_engineer: tuple[str] | None = self.get_contributors(
            "Vocal Engineer",
        )


@dataclass
class PlaylistsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API, videos/<VIDEOID> endpoint.If the params and
    headers are correctly specified, the API returns metadata of the available
    version of the (music) video, including video quality, video title, date,
    video artists, duration, etc."""

    uuid: str = field(repr=False)
    title: str
    number_of_tracks: int
    number_of_videos: int
    description: str
    created: Annotated[datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")]
    type: str
    public_playlist: bool
    url: str
    square_image: str  # UUID v4


class TidalResource:
    """Parent class to subclasses representing different TIDAL music
    service objects; e.g. Track, Album. This class is not meant to be
    instantiated itself: rather, its purpose is to pre-populate its
    subclasses with the `match_url` method."""

    def __init__(self, pattern: str | None = None, url: str | None = None):
        self.pattern = pattern
        self.url = url

    def match_url(self) -> int | str | None:
        _match: re.Match = re.match(self.pattern, self.url, re.IGNORECASE)
        try:
            _id: str = _match.groups()[0]
        except AttributeError:
            return None
        else:
            return _id


@dataclass
class TidalAlbum(TidalResource):
    """Class representing a TIDAL album. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/"
            r"(?:browse/)?album/(\d{2,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL album URL"
            raise ValueError(_msg)
        self.tidal_id = int(_id)
        _msg: str = f"TIDAL album ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


@dataclass
class TidalArtist(TidalResource):
    """Represent the concept of a TIDAL artist.

    Its main purpose is the __post_init__ checking process.
    """

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/"
            r"(?:browse/)?artist/(\d{2,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL album URL"
            raise ValueError(_msg)
        self.tidal_id = int(_id)
        _msg: str = f"TIDAL album ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


@dataclass
class TidalMix(TidalResource):
    """Represent the concept of a TIDAL mix.

    Its main purpose is the __post_init__ checking process.
    """

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/"
            r"(?:browse/)?mix/(\w{30})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL mix URL"
            raise ValueError(_msg)
        self.tidal_id = _id
        _msg: str = f"TIDAL mix ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


@dataclass
class TidalTrack(TidalResource):
    """Represent the concept of a TIDAL track.

    Its main purpose is the __post_init__ checking process.
    """

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/"
            r"(?:browse/)?(?:album/\d{5,9}/)?track/(\d{5,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL track URL"
            raise ValueError(_msg)
        self.tidal_id = int(_id)
        _msg: str = f"TIDAL track ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


@dataclass
class TidalPlaylist(TidalResource):
    """Represent the concept of a TIDAL playlist.

    Its main purpose is the __post_init__ checking process.
    """

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/(?:browse/)?playlist/"
            r"([0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12})(?:.*?)?"
        )

        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL playlist URL"
            raise ValueError(_msg)
        self.tidal_id = _id
        _msg: str = f"TIDAL playlist ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


@dataclass
class TidalVideo(TidalResource):
    """Represent the concept of a TIDAL video.

    Its main purpose is the __post_init__ checking process.
    """

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.|www\.)?tidal\.com/"
            r"(?:browse/)?video/(\d{7,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            _msg: str = f"'{self.url}' is not a valid TIDAL video URL"
            raise ValueError(_msg)

        self.tidal_id = int(_id)
        _msg: str = f"TIDAL video ID parsed from input: {self.tidal_id}"
        logger.info(_msg)


def match_tidal_url(input_str: str) -> TidalResource | None:
    """Attempt to match the `input_str` to either the URL of a track or an
    album in the Tidal API service. Returns None if `input_str` matches
    neither, otherwise a subclass of TidalResource corresponding to the
    parsed input_str type
    """
    resource_match: TidalResource | None = None
    tidal_resources: tuple[
        TidalResource,
        TidalResource,
        TidalResource,
        TidalResource,
        TidalResource,
        TidalResource,
    ] = (
        TidalTrack,
        TidalAlbum,
        TidalVideo,
        TidalPlaylist,
        TidalMix,
        TidalArtist,
    )
    for T in tidal_resources:
        try:
            resource_match = T(input_str)
        except ValueError as v:
            logger.debug(v)
            continue

        return resource_match


def download_artist_image(
    session: Session,
    artist: Artist,
    output_dir: Path,
    dimension: int = 320,
) -> Path | None:
    """Given a UUID that corresponds to a (JPEG) image on Tidal's servers,
    download the image file and write it as '{artist name}.jpeg'
    in the directory `output_dir`. Returns path to downloaded file"""
    _url: str = artist.picture_url(dimension)
    if _url is None:
        _msg: str = (
            f"Cannot download image for artist '{artist.name}', "
            "as Tidal supplied no URL for this artist's image."
        )
        logger.info(_msg)
        return None

    with session.get(url=_url, headers={"Accept": "image/jpeg"}) as r:
        if not r.ok:
            _msg: str = (
                "Could not retrieve data from Tidal resources/images URL "
                f"for artist '{artist.name}' due to error code: {r.status_code}"
            )
            logger.warning(_msg)
            logger.debug(r.reason)
            return None
        bytes_to_write: BytesIO = BytesIO(r.content)

    file_name: str = f"{artist.name.replace('..', '')}.jpg"
    output_file: Path | None = None
    if bytes_to_write is not None:
        output_file: Path = output_dir / file_name
        bytes_to_write.seek(0)
        output_file.write_bytes(bytes_to_write.read())
        bytes_to_write.close()
        _msg: str = (
            f"Wrote artist image JPEG for {artist} to "
            f"'{output_file.absolute()}'"
        )
        logger.info(_msg)
    return output_file



"""Represent the HLS manifests retrieved from TIDAL API."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

import m3u8
from requests import HTTPError, Session

if TYPE_CHECKING:
    from .models import VideosEndpointStreamResponseJSON

logger = logging.getLogger(__name__)


class TidalM3U8Error(Exception):
    """For the arising of exception in processing HLS's .m3u8 files."""


# https://github.com/globocom/m3u8#using-different-http-clients
class RequestsClient:
    """A custom class to pass to the m3u8.load() function."""

    def __init__(self, session: Session) -> None:
        self.session = session

    def download(
        self,
        url: str,
        timeout: int | None = None,
    ) -> tuple[str, str]:
        """Use self.session to GET `url`, returning response text and URL."""
        p: dict[str, None] = {k: None for k in self.session.params}
        with self.session.get(url=url, timeout=timeout, params=p) as response:
            return response.text, response.url


def playlister(
    session: Session,
    vesrj: VideosEndpointStreamResponseJSON | None,
) -> m3u8.M3U8:
    """Attempt to parse a VideosEndpointStreamResponseJSON into an m3u8.M3U8.

    Requires fetching HTTP(s) resources, so takes a requests.Session object
    as an argument. If exception occurs, raises TidalM3U8Error.
    """
    em_three_you_ate: m3u8.M3U8 | None = None
    if vesrj.manifest_mime_type == "application/vnd.tidal.emu":
        try:
            manifest: dict[str, str] = json.loads(vesrj.manifest_bytes)
        except json.decoder.JSONDecodeError as jde:
            _msg: str = (
                "Expected an HLS specification in JSON format "
                f"for video {vesrj.video_id}"
            )
            raise TidalM3U8Error(_msg) from jde
        else:
            mt: str | None = manifest.get("mimeType")
            url: str | None = manifest.get("urls", [None])[0]

        if (
            (mt is None)
            or (mt != "application/vnd.apple.mpegurl")
            or (url is None)
            or (".m3u8" not in url)
        ):
            _msg: str = (
                f"Manifest for video {vesrj.video_id}, video mode "
                f"{vesrj.video_quality} does not make available an "
                "M3U8 file"
            )
            raise TidalM3U8Error(_msg)

        download_params: dict[str, None] = {k: None for k in session.params}
        with session.get(url=url, params=download_params) as m3u8_response:
            try:
                m3u8_response.raise_for_status()
            except HTTPError as he:
                _msg: str = (
                    f"Could not retrieve variant streams from manifest for "
                    f"video {vesrj.video_id}, video mode {vesrj.video_quality}"
                )
                raise TidalM3U8Error(_msg) from he
            else:
                em_three_you_ate: m3u8.M3U8 | None = m3u8.M3U8(
                    content=m3u8_response.text,
                )

    return em_three_you_ate


def variant_streams(
    em3u8: m3u8.M3U8,
    session: Session,
    *,
    return_urls: bool = False,
) -> m3u8.M3U8 | list[str] | None:
    """By default, return the highest-bandwidth of em3u8.playlists as m3u8.M3U8 object.

    If return_urls, then the object's .files attribute is returned, which is a list of
    strings.
    N.b., if m3u8.is_variant is False, then return None as there are no variant streams.
    """
    if not em3u8.is_variant:
        return None

    playlist: m3u8.Playlist = max(
        em3u8.playlists,
        key=lambda p: p.stream_info.bandwidth,
    )
    if not return_urls:
        return playlist

    download_params: dict[str, None] = {k: None for k in session.params}
    with session.get(url=playlist.uri, params=download_params) as response:
        try:
            response.raise_for_status()
        except HTTPError as he:
            _msg: str = f"Could not retrieve media URLs from manifest {em3u8}"
            raise TidalM3U8Error(_msg) from he
        else:
            _m3u8: m3u8.M3U8 = m3u8.M3U8(content=response.text)

    return _m3u8.files
